# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:44:42+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


import numpy as np
import matplotlib.pyplot as plt


class Chaos_map(object):

    """Chaos_map

    Chaos_map is used to build the chaotic map according to the give name, the level and dimensions of the Searchspace.
    See Chaotic_optimization for more info. Th chaotic map is matrix of shape: (level, dimension).

    Attributes
    ----------
    level : int
        Chaotic level corresponds to the number of iteration of the chaotic map
    dimension : int
        Dimension of the Searchspace.
    map_kwargs : dict
        Keyword arguments associated to the selected map.
    map : np.array(dtype=np.array(dtype=float))
        Contains the chaotic map of size level*dimension.

    See Also
    --------
    Chaotic_optimization : Chaos map is used here.
    """

    def __init__(self, vectors, params):

        self.vectors = vectors
        self.params = params
        self.map = np.zeros([self.vectors, self.params])

    def __add__(self, map):
        assert (
            self.params == map.params
        ), f"Error, maps must have equals `params`, got {self.params} and {map.params}"

        new_map = Chaos_map(self.vectors + map.vectors, self.params)

        new_map.map = np.append(self.map, map.map)
        np.random.shuffle(new_map.map)

        return new_map


class Henon(Chaos_map):
    """Henon chaotic map

    .. math::

        \\smash{
        \\begin{cases}
        \\begin{cases}
        x_{n+1} = 1 - a x_n^2 + y_n\\\\
        y_{n+1} = b x_n.
        \\end{cases}\\\\
        map = \\frac{y-min(y)}{max(y)-min(y)}
        \\end{cases}}

    Parameters
    ----------
    vectors : int
        Map size
    params : int
        Number of dimensions
    a : float, default=1.4020560
        Henon map parameter. Has an influence on the chaotic, intermittent or periodicity behaviors.
    b : float, default=0.305620406
        Henon map parameter. Has an influence on the chaotic, intermittent or periodicity behaviors.

    Attributes
    ----------
    map : numpy.ndarray
        Chaos map of size (vectors,param)

    """

    def __init__(self, vectors, params, a=1.4020560, b=0.305620406):

        super().__init__(vectors, params)

        self.a = a
        self.b = b

        # Initialization
        y = np.zeros([self.vectors, self.params])
        x = np.random.random(self.params)

        for i in range(1, self.vectors):

            # y_{k+1} = x_{k}
            y[i, :] = b * x

            # x_{k+1} = a.(1-x_{k}^2) + b.y_{k}
            x = 1 - a * x ** 2 + y[i - 1, :]

        # Min_{params}(y_{params,vectors})
        alpha = np.amin(y, axis=0)

        # Max_{params}(y_{params,vectors})
        beta = np.amax(y, axis=0)

        self.map = (y - alpha) / (beta - alpha)


class Kent(Chaos_map):
    """Kent chaotic map

    .. math::

        \\smash{
        \\begin{cases}
        x_{n+1} =
        \\begin{cases}
        \\frac{x_n}{\\beta} \\quad 0 < x_{n}  \\leq \\beta \\\\
        \\frac{1-x_n}{1-\\beta} \\quad \\beta < x_{n} \\leq 1
        \\end{cases}\\\\
        map=x
        \\end{cases}}

    Parameters
    ----------
    vectors : int
        Map size
    params : int
        Number of dimensions
    beta : float, default=0.8
        Kent map parameter. Has an influence on the chaotic, intermittent, convergence, or periodicity behaviors.

    Attributes
    ----------
    map : numpy.ndarray
        Chaos map of size (vectors,param)

    """

    def __init__(self, vectors, params, beta=0.8):
        super().__init__(vectors, params)

        self.beta = beta

        self.map[0, :] = np.random.random(params)

        for i in range(1, vectors):
            self.map[i, :] = np.where(
                self.map[i - 1, :] < beta,
                self.map[i - 1, :] / beta,
                (1 - self.map[i - 1, :]) / (1 - beta),
            )


class Logistic(Chaos_map):
    """Logistic chaotic map

    .. math::

        \\smash{
        \\begin{cases}
        x_{n+1} = \\mu x_n(1-x_n)\\\\
        map=x
        \\end{cases}}

    Parameters
    ----------
    vectors : int
        Map size
    params : int
        Number of dimensions
    mu : float, default=3.57
        Logistic map parameter. Has an influence on the chaotic, intermittent, convergence, or periodicity behaviors.

    Attributes
    ----------
    map : numpy.ndarray
        Chaos map of size (vectors,param)

    """

    def __init__(self, vectors, params, mu=3.57):

        super().__init__(vectors, params)

        self.mu = mu

        self.map[0, :] = np.random.random(params)

        for i in range(1, vectors):
            self.map[i, :] = mu * self.map[i - 1, :] * (1 - self.map[i - 1, :])


class Tent(Chaos_map):
    """Tent chaotic map

    .. math::

        \\smash{
        \\begin{cases}
        x_{n+1} =
        \\begin{cases}
        \\mu x_n     \\quad x_n < \\frac{1}{2} \\\\
        \\mu (1-x_n) \\quad \\frac{1}{2} \\leq x_n
        \\end{cases}\\\\
        map = x
        \\end{cases}}

    Parameters
    ----------
    vectors : int
        Map size
    params : int
        Number of dimensions
    Tent : float, default=1.9999999999
        Logistic map parameter. Has an influence on the chaotic, intermittent, convergence, or periodicity behaviors.

    Attributes
    ----------
    map : numpy.ndarray
        Chaos map of size (vectors,param)

    """

    def __init__(self, vectors, params, mu=2 - 1e-10):

        super().__init__(vectors, params)

        self.mu = mu

        self.map[0, :] = np.random.random(params)

        for i in range(1, vectors):
            self.map[i, :] = np.where(
                self.map[i - 1, :] < 0.5,
                self.mu * self.map[i - 1, :],
                self.mu * (1 - self.map[i - 1, :]),
            )


class Random(Chaos_map):
    def __init__(self, vectors, params):
        super().__init__(vectors, params)

        self.map = np.random.random((vectors, params))


chaos_map_name = {
    "henon": Henon,
    "logistic": Logistic,
    "kent": Kent,
    "tent": Tent,
    "random": Random,
}



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:44:16+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


import zellij.utils.progress_bar as pb
from abc import abstractmethod
import os
import numpy as np
import pandas as pd
import enlighten

import logging

logger = logging.getLogger("zellij.Meta")


class Metaheuristic(object):

    """Metaheuristic

    Metaheuristic is a core object which defines the structure of a metaheuristic in Zellij.
    It is an abtract class.

    Attributes
    ----------
    loss_func : LossFunc
        Loss function to optimize. must be of type :math:`f(x)=y` or :math:`f(x)=results,model`
        See :ref:`lf` for more information.

    search_space : Searchspace
        :ref:`sp` object containing bounds of all decision variables.

    f_calls : int
        Maximum number of calls to loss_func.

    save : boolean, optional
        If True save results into a file

    H : Fractal, optional
        If a :ref:`frac` is given, allows to use it.

    verbose : boolean, default=True
        Activate or deactivate the progress bar.

    See Also
    --------
    :ref:`lf` : Parent class for a loss function.
    :ref:`sp` : Defines what a search space is in Zellij.
    """

    def __init__(self, loss_func, search_space, f_calls, verbose=True):

        ##############
        # PARAMETERS #
        ##############

        self.loss_func = loss_func
        self.search_space = search_space
        self.f_calls = f_calls

        self.verbose = verbose

        #############
        # VARIABLES #
        #############

        # Modify labels in loss func according to SearchSpace labels
        self.loss_func.labels = self.search_space.labels
        # Index of the historic in loss function.
        self.lf_idx = len(self.loss_func.all_scores)

        if self.verbose:
            self.manager = enlighten.get_manager()
        else:
            self.manager = enlighten.get_manager(stream=None, enabled=False)

        self.main_pb = False

    def build_bar(self, total):
        """build_bar(total)

        build_bar is a method to build a progress bar.
        It is a purely aesthetic feature to get info on the execution.
        You can deactivate it, with `verbose=False`.

        Parameters
        ----------
        total : int
            Length of the progress bar.

        """

        if self.verbose:
            if (not hasattr(self.manager, "zellij_first_line")) or (
                hasattr(self.manager, "zellij_first_line")
                and not self.manager.zellij_first_line
            ):

                self.main_pb = True
                self.manager.zellij_first_line = True
                self.best_pb = pb.best_counter(self.manager)
                (
                    self.calls_pb_explor,
                    self.calls_pb_exploi,
                    self.calls_pb_pending,
                ) = pb.calls_counter(self.manager, self.f_calls)

                self.loss_func.manager = self.manager

            else:
                self.main_pb = False
                self.best_pb = False
                self.calls_pb_explor = False
                self.calls_pb_exploi = False
                self.calls_pb_pending = False

        self.meta_pb = pb.metaheuristic_counter(
            self.manager, total, self.__class__.__name__
        )

    def update_main_pb(self, nb, explor=True, best=False):
        """update_main_pb(nb, explor=True, best=False)

        Update the main progress bar with a certain number.

        Parameters
        ----------
        nb : int
            Length of the update. e.g. if the progress bar measure the number of iterations,
            at each iteration `nb=1`.
        explor : bool, default=True
            If True the color associated to the update will be blue. Orange, otherwise.
        best : bool default=False
            If True the score of the current solution will be displayed.

        """
        if self.main_pb and self.verbose:
            if best:
                self.best_pb.update()
            if explor:
                self.calls_pb_explor.update_from(self.calls_pb_pending, nb)
            else:
                self.calls_pb_exploi.update_from(self.calls_pb_pending, nb)

    def pending_pb(self, nb):
        """pending_pb(nb)

        Update the progress bar with a pending property (white). This update will be replaced when using
        `update_main_pb`.

        Parameters
        ----------
        nb : type
            Length of the pending objects.

        """
        if self.main_pb and self.verbose:
            self.calls_pb_pending.update(nb)

    def close_bar(self):
        """close_bar()

        Delete the progress bar. (must be executed at the end of `run` method)

        """
        if self.main_pb and self.verbose:
            self.best_pb.close()
            self.calls_pb_pending.close()

            self.main_pb = False
            self.manager.zellij_first_line = False

        self.meta_pb.close()

    @abstractmethod
    def run(self):
        """run()

        Abstract method, describe how to run a metaheuristic

        """
        pass

    def show(self, filepath="", save=False):
        """show(filepath="", save=False)

        Basic plots for all metaheuristic. uses :ref:`sp` plotting.

        Parameters
        ----------
        filepath : string, default=""
            If a filepath to a file containing points is given, it will read those points and plot them.
        save : bool, default=False
            If true, it saves the plots.

        """

        if filepath:
            all = os.path.join(filepath, "outputs", "all_evaluations.csv")

            all_data = pd.read_table(all, sep=",", decimal=".")
            all_scores = all_data["loss"].to_numpy()
        else:
            all_data = self.loss_func.all_solutions
            all_scores = np.array(self.loss_func.all_scores)

        self.search_space.show(
            all_data, all_scores, save, self.loss_func.plots_path
        )

        return all_data, all_scores



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:46:01+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


import random
import time
import enlighten


def best_counter(manager):
    return manager.counter(desc="Improvements", unit="solutions", leave=False)


def calls_counter(manager, total):

    bar_format = (
        "{desc}{desc_pad}{percentage:3.0f}%|{bar}| "
        + "Pending:{count_0:{len_total}d} "
        + "Explor:{count_2:{len_total}d} "
        + "Exploi:{count_1:{len_total}d}"
    )

    pending = manager.counter(
        total=total,
        desc="Loss calls",
        unit="calls",
        color="white",
        bar_format=bar_format,
        leave=False,
    )
    exploitation = pending.add_subcounter("orange")
    exploration = pending.add_subcounter("cyan")

    return exploration, exploitation, pending


def metaheuristic_counter(manager, total, name):
    bar_format = "{desc}{desc_pad}{percentage:3.0f}%|{bar}| "

    counter = manager.counter(
        total=total,
        desc=f"   {name}",
        unit="calls",
        color="white",
        bar_format=bar_format,
        leave=False,
    )
    return counter


def calls_counter_inside(manager, total):
    bar_format = "{desc}{desc_pad}{percentage:3.0f}%|{bar}| "

    counter = manager.counter(
        total=total,
        desc=f"      Evaluating",
        unit="calls",
        color="white",
        bar_format=bar_format,
        leave=False,
    )
    return counter


def best_found(manager, score):
    status_bar = manager.status_bar(
        "      Current score:{:.3f} | Best score:{:.3f}".format(score, score),
        color="white",
        leave=False,
    )
    return status_bar



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:44:48+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


import numpy as np
from abc import abstractmethod
import matplotlib.pyplot as plt
from scipy.spatial.distance import cdist

import logging

logger = logging.getLogger("zellij.spokedart")


def randomMuller(n, m, surface=False):
    u = np.random.normal(0, 1, (n, m))
    d = np.linalg.norm(u, axis=1, keepdims=True)
    p = u / d

    if surface:
        return p
    else:
        return np.random.random(n) ** (1 / m) * p


def interSegmentCircle(A, B, C, R):

    dist = np.linalg.norm(B - A)

    BmA = B - A
    CmA = C - A

    a = np.sum(BmA ** 2)
    b = -2 * np.sum(BmA * CmA)
    c = np.sum(CmA ** 2) - R ** 2

    delta = b ** 2 - 4 * a * c

    intersect = np.empty((0, len(C)))

    if delta < 0:
        p1, p2 = (None, None)
    elif delta > 0:
        p1, p2 = (
            A + ((-b + np.sqrt(delta)) / (2 * a)) * BmA,
            A + ((-b - np.sqrt(delta)) / (2 * a)) * BmA,
        )

        Ap1, p1B = np.linalg.norm(p1 - A), np.linalg.norm(B - p1)
        Ap2, p2B = np.linalg.norm(p2 - A), np.linalg.norm(B - p2)
        if dist == Ap1 + p1B:
            intersect = np.append(intersect, [p1], axis=0)
        if dist == Ap2 + p2B:
            intersect = np.append(intersect, [p2], axis=0)

        # intersect = np.append(intersect, [p1, p2], axis=0)

    else:
        p1 = A + (-b / (2 * a)) * BmA
        dot = np.dot(BmA, BmA)
        dotc1 = np.dot(BmA, p1 - A)
        if dotc1 >= 0 and dotc1 <= dot:
            intersect = np.append(intersect, [p1], axis=0)

    return intersect


def interLineHyperplane(A, B, X, Y):
    """interLineHyperplane(A, B, X, Y)

    Compute the intersection between line AB and the plane separating X and Y.

    Parameters
    ----------
    A : type
        First point from line
    B : type
        Second point from line
    X : type
        First point under plane
    Y : type
        Second point above plane

    Returns
    -------
    bool
        if true the intersection is found, else no
    np.ndarray
        point intersection

    """

    if np.all(X == Y):
        return False, -1
    else:
        # line vector
        BmA = B - A

        # Normal vector for bisecting plane (plane at the middle of X and Y)
        YmX = Y - X
        v = YmX / np.linalg.norm(YmX)
        xmid = (X + Y) / 2

        if np.dot(BmA, v) == 0:
            return False, -1
        else:
            d = np.dot(xmid - A, v) / np.dot(BmA, v)
            intersection = A + BmA * d

            # C = A + (B-A)*t, use parametric and symetric equation of a line: (c0-a0)/(b0-a0) = (c1-a1)/(b1-a1)... = t, be caraful when (bi -ai) = 0, t doesn't exists for dim i.
            maxidx = np.argmax(
                BmA != 0
            )  # argmax stops at the first occurence of True. At least one value from BmA must be != 0, otherwise, A==B which is not possible.
            t = (intersection[maxidx] - A[maxidx]) / BmA[maxidx]

            # not on the half line
            if t < 0:
                return False, -1
            else:
                return True, intersection


class HalfLine(object):
    def __init__(self, A, B):
        self.A = A
        self.B = B

        self.v = B - A

    def point(self, d):
        return self.A + self.v * d


class Hyperplane(object):
    def __init__(self, X, Y):

        self.cellX = X
        self.cellY = Y

        self.X = X.seed
        self.Y = Y.seed

        assert not np.all(self.X == self.Y), logger.error(
            f"Can't build a bisecting plane between two identical points, got {X.seed} and {Y.seed}"
        )

        YmX = self.Y - self.X

        self.xmid = (self.X + self.Y) / 2
        self.v = YmX / np.linalg.norm(YmX)

    def intersection(self, line):
        # line is on the plane /!\ impossible in our case
        # if np.dot(line.v, self.v) == 0:
        #     return False, -1
        # else:

        d = np.dot(self.xmid - line.A, self.v) / np.dot(line.v, self.v)

        # not on the half line
        if d < 0:
            return False, -1
        else:
            return True, line.point(d)


class SpokeDart(object):
    def __init__(self, dim, seeds, r, reject=12):

        ##############
        # PARAMETERS #
        ##############

        self.seeds = seeds
        self.radius = r
        self.m = reject
        self.dim = dim

    def run(self, max):

        s = np.random.random(self.dim)

        S = np.empty((0, self.dim))
        P = np.array([s])
        N = []

        while len(P) > 0 and len(S) < max:

            s = P[0]
            P = P[1:]

            if len(S) > 0:
                N = list(self.collect_neighbors(s, S))

            reject = 0
            while reject < self.m and len(S) < max:
                a, b = self.randomSpoke(s)

                if len(N) > 0:
                    covered, a, b = self.trimSpoke(a, b, S[N])
                else:
                    covered = False

                if covered:
                    reject += 1
                else:
                    u = b - a

                    ulength = np.linalg.norm(u)
                    if ulength > 0:
                        rlength = np.random.uniform(0, ulength)
                        v = u * rlength / ulength

                        sp = a + v

                    else:
                        sp = a

                    if np.any(sp > np.array([1] * self.dim)) or np.any(
                        sp < np.array([0] * self.dim)
                    ):
                        reject += 1
                    else:
                        S = np.append(S, [sp[:]], axis=0)
                        N.append(len(S) - 1)
                        P = np.append(P, [sp[:]], axis=0)

                        reject = 0

        return S

    def collect_neighbors(self, s, samples):
        dist = cdist([s], samples)[0]
        mask = dist < 3 * self.radius
        return mask.nonzero()[0]

    def randomSpoke(self, s):
        p = randomMuller(1, self.dim, surface=True)[0]
        pclose = p * self.radius + s
        pfar = p * 2 * self.radius + s
        return pclose, pfar

    def trimSpoke(self, a, b, neighbors):
        trims = np.empty((0, self.dim))
        covered = False
        i = 0
        while i < len(neighbors) and not covered:
            n = neighbors[i]
            i += 1

            if np.linalg.norm(n - a) <= self.radius:
                covered = True
            else:
                res = interSegmentCircle(a, b, n, self.radius)
                trims = np.append(trims, res, axis=0)

        if covered:
            return covered, -1, -1
        elif len(trims) > 0:
            dist = np.linalg.norm(trims - a, axis=1)
            minidx = np.argmin(dist)
            return covered, a, trims[minidx]
        else:
            return covered, a, b



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:44:13+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


import numpy as np
import os
import shutil
from abc import abstractmethod
import enlighten
import zellij.utils.progress_bar as pb

import logging

logger = logging.getLogger("zellij.loss")

try:
    from mpi4py import MPI
except ImportError as err:
    logger.info(
        "To use MPILoss object you need to install mpi4py and an MPI distribution\n\
    You can use: pip install zellij[MPI]"
    )


class LossFunc(object):

    """LossFunc

    LossFunc allows to wrap function of type :math:`f(x)=y`. With :math:`x` a set of hyperparameters.
    However, **Zellij** supports alternative pattern: :math:`f(x)=results,model` for example.
    Where:

    * :math:`results` can be a `list <https://docs.python.org/3/tutorial/datastructures.html#more-on-lists>`_ or a `dictionary <https://docs.python.org/3/tutorial/datastructures.html#dictionaries>`_. The first element of the list must be the loss value. If the return is a dictionary, the loss value must have the key *"score"*.
    * :math:`model` is optionnal, it is an object with a *save()* method. (e.g. a neural network from Tensorflow)

    You must wrap your function so it can be used in Zellij by adding several features,\
     such as calls count, saves, parallelization, historic...

    Attributes
    ----------
    model : function
        Function of type :math:`f(x)=y` or :math:`f(x)=results,model. :math:`x` must be a solution. A solution can be a list of float, int... It can also be of mixed types...
    best_score : float
        Best score found so far.
    best_sol : list
        Best solution found so far.
    best_argmin : int
        Index of the best solution found so far.
    all_scores : float
        Historic of all evaluated scores.
    all_solutions : float
        Historic of all evaluated solutions.
    calls : int
        Number of loss function calls

    See Also
    --------
    Loss : Wrapper function
    MPILoss : Distributed version of LossFunc
    SerialLoss : Basic version of LossFunc
    """

    def __init__(self, model, save=False, verbose=True):

        """__init__(model, save=False)

        Parameters
        ----------
        model : function
            Function of type :math:`f(x)=y` or :math:`f(x)=results,model. :math:`x` must be a solution. A solution can be a list of float, int... It can also be of mixed types...
        save : string
            Filename where to save the best found model and the historic of the loss function.
            Only one model is saved for memory issues. Be carefull, if you want to save a model,
            the object that you loss function returns, must have a "save" method with a filename parameter. (e.g. model.save(filename)).

        """
        ##############
        # PARAMETERS #
        ##############

        self.model = model
        self.save = save
        self.verbose = verbose
        #############
        # VARIABLES #
        #############

        self.best_score = float("inf")
        self.best_sol = None
        self.best_argmin = None

        self.all_scores = []
        self.all_solutions = []

        self.calls = 0
        # Must be private, à voir
        self.new_best = False

        self.labels = []

        if isinstance(self.save, str):
            self.folder_name = self.save
        else:
            self.folder_name = f"{self.model.__class__.__name__}_zlj_save"

        self.outputs_path = ""
        self.model_path = ""
        self.plots_path = ""
        self.loss_file = ""

        self.file_created = False

        if self.verbose:
            self.manager = enlighten.get_manager()
        else:
            self.manager = enlighten.get_manager(stream=None, enabled=False)

    def build_bar(self, total):
        """build_bar(total)

        build_bar is a method to build a progress bar.
        It is a purely aesthetic feature to get info on the execution.
        You can deactivate it, with `verbose=False`.

        Parameters
        ----------
        total : int
            Length of the progress bar.

        """

        if self.verbose:
            self.lf_pb = pb.calls_counter_inside(self.manager, total)
            self.best_pb = pb.best_found(self.manager, self.best_score)

    def close_bar(self):
        """close_bar()

        Delete the progress bar.

        """
        if self.verbose:
            self.lf_pb.close()
            self.best_pb.close()

    @abstractmethod
    def _save_model(self, *args):
        """ _save_model()

        Private abstract method to save a model. Be carefull, to be exploitable, the initial loss func must be of form :math:`f(x) = (y, model)`\
         `y` are the results of the evaluation of `x` by `f`. `model` is optional, if you want to save the best model found (e.g. a neural network)\
         you can return the model. However the model must have a "save" method with a filename. (e.g. model.save(filename)).

        """
        pass

    @abstractmethod
    def __call__(self, X, **kwargs):
        pass

    def _create_file(self, x, *args):

        """create_file(x, *args)

        Create a save file:

        Structure:

            foldername
            | model # if sav = True in LossFunc, contains model save
              | model_save
            | outputs # Contains loss function outputs
              | file_1.csv
              | ...
            | plots # if save = True while doing .show(), contains plots
              | plot_1.png
              | ...

        Parameters
        ----------
        solution : list
            Needs a solution to determine the header of the save file

        *args : list[label]
            list of additionnal labels to add before the score/evaluation of a point.

        """

        # Create a valid folder
        try:
            os.mkdir(self.folder_name)
            created = False
        except FileExistsError as error:
            created = True

        i = 0
        while created:
            try:
                nfolder = f"{self.folder_name}_{i}"
                os.mkdir(nfolder)
                created = False
                logger.warning(
                    f"WARNING: Folder {self.folder_name} already exists, results will be saved at {nfolder}"
                )
                self.folder_name = nfolder
            except FileExistsError as error:
                i += 1

        # Create ouputs folder
        self.outputs_path = os.path.join(self.folder_name, "outputs")
        os.mkdir(self.outputs_path)

        self.model_path = os.path.join(self.folder_name, "model")
        os.mkdir(self.model_path)

        self.plots_path = os.path.join(self.folder_name, "plots")
        os.mkdir(self.plots_path)

        # Additionnal header for the outputs file
        if len(args) > 0:
            suffix = "," + ",".join(str(e) for e in args)
        else:
            suffix = ""

        # Create base outputs file for loss func
        self.loss_file = os.path.join(self.outputs_path, "all_evaluations.csv")

        # Determine header
        if len(self.labels) != len(x):
            logger.warning(
                "WARNING: Labels are of incorrect size, it will be replaced in the save file header"
            )
            for i in range(len(x)):
                self.labels.append(f"attribute{i}")

        with open(self.loss_file, "w") as f:
            f.write(
                ",".join(str(e) for e in self.labels) + ",loss" + suffix + "\n"
            )

        logger.info(
            f"INFO: Results will be saved at: {os.path.abspath(self.folder_name)}"
        )

        self.file_created = True

    def _save_file(self, x, y, **kwargs):

        """_save_file(x, y, **kwargs)

        Private method to save informations about an evaluation of the loss function.

        Parameters
        ----------
        x : list
            Solution to save.
        y : list
            Evaluation of the solution by the loss function;
        filename : str
            Name of the file created by a Metaheuristic, where to save informations.
        **kwargs : dict, optional
            Other informations to save after the score.
        """

        if not self.file_created:
            self._create_file(x, *list(kwargs.keys()))

        # Determine if additionnal contents must be added to the save
        if len(kwargs) > 0:
            suffix = "," + ",".join(str(e) for e in kwargs.values())
        else:
            suffix = ""

        # Save a solution and additionnal contents
        with open(self.loss_file, "a+") as f:
            f.write(",".join(str(e) for e in x) + "," + str(y) + suffix + "\n")

    # Save best found solution
    def _save_best(self, x, y):
        """_save_best(x, y)

        Save point x with score y, and verify if this point is the best found so far.

        Parameters
        ----------
        x : list
            Set of hyperparameters (a solution)
        y : {float, int}
            Loss value (score) associated to x.

        """

        # historic
        self.all_solutions.append(list(x)[:])
        self.all_scores.append(y)

        # Save best
        if y < self.best_score:
            self.best_score = y
            self.best_sol = list(x)[:]
            self.best_argmin = len(self.all_scores)

            self.new_best = True

        if self.verbose:
            self.lf_pb.update(1)
            self.best_pb.update(
                "      Current score:{:.3f} | Best score:{:.3f}".format(
                    y, self.best_score
                ),
                color="white",
            )

    def _build_return(self, r):
        """_build_return(r)

        This method builds a unique return according to the outputs of the loss function

        Parameters
        ----------
        r : {list, float, int}
            Returns of the loss function

        Returns
        -------
        rd : dict
            Dictionnary mapping outputs from the loss function

        model : object
            Model object with a 'save' method

        """

        rd = {}

        # Separate results and model
        if isinstance(r, tuple):
            if len(r) > 1:
                results, model = r
            else:
                results, model = r, False
        else:
            results, model = r, False

        # Separate results
        if isinstance(results, int) or isinstance(results, float):
            rd["score"] = results
        elif isinstance(results, dict):
            rd = results
            rd["score"] = r.values()[0]
        elif isinstance(results, list):
            rd["score"] = results[0]
            for i, j in enumerate(results):
                label = f"return{i}"
                rd[label] = j

        return rd, model

    def reset(self):
        """reset()

        Reset all attributes of `LossFunc` at their initial values.

        """

        self.best_score = float("inf")
        self.best_sol = None
        self.best_argmin = None

        self.all_scores = []
        self.all_solutions = []

        self.calls = 0
        # Must be private, à voir
        self.new_best = False

        self.labels = []

        if isinstance(self.save, str):
            self.folder_name = self.save
        else:
            self.folder_name = f"{self.model.__class__.__name__}_zlj_save"

        self.outputs_path = ""
        self.model_path = ""
        self.plots_path = ""
        self.loss_file = ""

        self.file_created = False

        if self.verbose:
            self.manager = enlighten.get_manager()
        else:
            self.manager = enlighten.get_manager(stream=None, enabled=False)


class FDA_loss_func:

    """FDA_loss_func

    FDA_loss_func allows to wrap function of type f(x)=(y, model), so it can be used in by the Fractal Decomposition Algorithm.

    Must be modified
    """

    def __init__(self, model, H, sp):

        ##############
        # PARAMETERS #
        ##############

        self.loss_func = model
        self.H = H
        self.search_space = sp

    @property
    def calls(self):
        return self.loss_func.calls

    @property
    def save(self):
        return self.loss_func.save

    @property
    def best_score(self):
        return self.loss_func.best_score

    @property
    def best_sol(self):
        return self.loss_func.best_sol

    @property
    def all_scores(self):
        return self.loss_func.all_scores

    @property
    def all_solutions(self):
        return self.loss_func.all_solutions

    @property
    def new_best(self):
        return self.loss_func.new_best

    @property
    def labels(self):
        return self.loss_func.labels

    @labels.setter
    def labels(self, v):
        self.loss_func.labels = v

    @property
    def folder_name(self):
        return self.loss_func.folder_name

    @property
    def outputs_path(self):
        return self.loss_func.outputs_path

    @property
    def model_path(self):
        return self.loss_func.model_path

    @property
    def plots_path(self):
        return self.loss_func.plots_path

    @property
    def loss_file(self):
        return self.loss_func.loss_file

    @property
    def file_created(self):
        return self.loss_func.file_created

    @file_created.setter
    def file_created(self, v):
        self.loss_func.file_created = v

    def __call__(self, X, **kwargs):
        res = self.loss_func(X, **kwargs)
        X_c = self.search_space.convert_to_continuous(X)
        self.H.add_point(res, X_c)
        return res


class MPILoss(LossFunc):

    """MPILoss

    MPILoss allows to wrap function of type f(x)=(y, model). MPILoss adds method to distribute dynamically the evaluation of multiple solutions.
    It does not distribute the original loss function itself

    Attributes
    ----------
    comm : MPI_COMM_WORLD
        All created processes and their communication context are grouped in comm.
    status : MPI_Status
        Data structure containing information about a received message.
    rank : int
        Process rank
    p : int
        comm size
    master : boolean
        If True the process is the master, else it is the worker.

    Methods
    -------
    __call__(X, filename='', **kwargs)
        Evaluate a list X of solutions with the original loss function.

    worker()
        Initialize a worker.

    stop()
        Stops all the workers and master.

    _save_model(score, source)
        See LossFunc, save a model according to its score and the worker rank.

    See Also
    --------
    Loss : Wrapper function
    LossFunc : Inherited class
    SerialLoss : Basic version of LossFunc
    """

    def __init__(self, model, save=False, verbose=True):

        """__init__(model, save=False)

        Initialize MPI variables. For more info, see LossFunc.

        """

        super().__init__(model, save, verbose)

        #################
        # MPI VARIABLES #
        #################

        try:
            self.comm = MPI.COMM_WORLD
            self.status = MPI.Status()
            self.rank = self.comm.Get_rank()
            self.p = self.comm.Get_size()
        except Exception as err:
            logger.error(
                "To use MPILoss object you need to install mpi4py and an MPI distribution\n\
            You can use: pip install zellij[Parallel]"
            )

            raise err

        # Master or worker process
        self.master = self.rank == 0

        if self.master:
            if os.path.exists("tmp_wks"):
                shutil.rmtree("tmp_wks")
            os.makedirs("tmp_wks")
        else:
            self.worker()

    def __call__(self, X, label=[], **kwargs):

        """__call__(model, save_model='')

        Evaluate a list X of solutions with the original loss function.

        Parameters
        ----------
        X : list
            List of solutions to evaluate. be carefull if a solution is a list X must be a list of lists.
        **kwargs : dict, optional
            Additionnal informations to save before the score.

        Returns
        -------
        res : list
            Return a list of all the scores corresponding to each evaluated solution of X.

        """

        logger.info("Master Starting")

        assert self.p > 1, "n_process must be > 1"

        self.build_bar(len(X))

        self.new_best = False

        res = [None] * len(X)
        send_history = [-1] * (self.p)

        nb_send = 0

        # Send a solution to all available processes
        while nb_send < len(X) and nb_send < (self.p - 1):

            logger.debug(f"MASTER {self.rank} sending to {nb_send}")

            self.comm.send(dest=nb_send + 1, tag=0, obj=X[nb_send])
            send_history[nb_send + 1] = nb_send
            nb_send += 1

        # Dynamically send and receive solutions and results to and from workers
        nb_recv = 0
        while nb_send < len(X):

            logger.debug(f"MASTER {self.rank} receiving | {nb_recv} < {len(X)}")

            msg, others = self.comm.recv(
                source=MPI.ANY_SOURCE, tag=0, status=self.status
            )
            source = self.status.Get_source()

            logger.debug(f"MASTER {self.rank} received from {source}")

            res[send_history[source]] = msg

            if self.save:
                # Save model into a file if it is better than the best found one
                self._save_model(msg, source)

                # Save score and solution into a file
                self._save_file(
                    X[send_history[source]],
                    res[send_history[source]],
                    **others,
                    **kwargs,
                )

            # Save score and solution into the object
            self._save_best(X[send_history[source]], res[send_history[source]])

            nb_recv += 1
            self.calls += 1

            logger.debug(f"MASTER {self.rank} sending to {nb_send}")

            self.comm.send(dest=source, tag=0, obj=X[nb_send])
            send_history[source] = nb_send

            nb_send += 1

        # Receive last results from workers
        while nb_recv < len(X):

            logger.debug(
                f"MASTER {self.rank} last receiving | {nb_recv} < {len(X)}"
            )

            msg, others = self.comm.recv(
                source=MPI.ANY_SOURCE, tag=0, status=self.status
            )
            source = self.status.Get_source()

            logger.debug(f"MASTER {self.rank} received from {source}")

            nb_recv += 1
            self.calls += 1

            res[send_history[source]] = msg

            if self.save:
                # Save model into a file if it is better than the best found one
                self._save_model(msg, source)

                # Save score and solution into a file
                self._save_file(
                    X[send_history[source]],
                    res[send_history[source]],
                    **others,
                    **kwargs,
                )

            # Save score and solution into the object
            self._save_best(X[send_history[source]], res[send_history[source]])

        self.close_bar()

        logger.info("Master ending")

        return res

    def worker(self):

        """worker()

        Initialize worker. Whilte it does not receive a stop message, a worker will wait for a solution to evaluate.

        """

        logger.info(f"Worker{self.rank} starting")

        stop = True

        while stop:

            logger.debug(f"WORKER {self.rank} receving")
            msg = self.comm.recv(source=0, tag=0, status=self.status)

            if msg != None:

                logger.debug(f"WORKER {self.rank} evaluating: {msg}")

                res, trained_model = self._build_return(self.model(msg))

                score, others = res["score"], res

                # Verify if a model is returned or not
                # Save the model using its save method
                if trained_model and self.save:
                    if hasattr(trained_model, "save") and callable(
                        getattr(trained_model, "save")
                    ):
                        worker_path = os.path.join(
                            "tmp_wks", f"worker{self.rank}"
                        )
                        os.system(f"rm -rf {worker_path}")
                        trained_model.save(worker_path)
                    else:
                        logger.error(
                            "Model/loss function does not have a method called `save`"
                        )
                        exit()

                # Send results
                logger.debug(f"WORKER {self.rank} sending {score}")

                self.comm.send(dest=0, tag=0, obj=[score, others])
            else:
                stop = False

        logger.info(f"Worker{self.rank} ending")

    def stop(self):

        """stop()

        Send a stop message to all workers.

        """

        for i in range(1, self.p):
            self.comm.send(dest=i, tag=0, obj=None)

        shutil.rmtree("tmp_wks", ignore_errors=True)

    def _save_model(self, score, source):

        """ _save_model()

        Private method to save a model. Be carefull, to be exploitable, the initial loss func must be of form f(x) = (y, model)\
         y is the results of the evaluation of x by f. model is optional, if you want to save the best found model (e.g. a neural network)\
         you can return the model. However the model must have a "save" method (e.g. model.save(filename)).

         score : int
            Score corresponding to the source file
        source : int
            Worker rank which evaluate a solution and return score

        """

        # Save model into a file if it is better than the best found one
        if score < self.best_score:

            master_path = ave_path = os.path.join(
                self.model_path, f"{self.model.__class__.__name__}_best"
            )
            worker_path = os.path.join("tmp_wks", f"worker{source}")

            if os.path.isdir(worker_path):
                os.system(f"rm -rf {master_path}")
                os.system(f"cp -rf {worker_path} {master_path}")


class SerialLoss(LossFunc):

    """SerialLoss

    SerialLoss allows to wrap function of type f(x)=(y, model). SerialLoss adds methods to save and evaluate the original loss function.

    Methods
    -------

    __call__(X, filename='', **kwargs)
        Evaluate a list X of solutions with the original loss function.

    _save_model(score, source)
        See LossFunc, save a model according to its score and the worker rank.

    See Also
    --------
    Loss : Wrapper function
    LossFunc : Inherited class
    MPILoss : Distributed version of LossFunc
    """

    def __init__(self, model, save=False, verbose=True):

        """__init__(model, save=False)

        Initialize SerialLoss.

        """

        super().__init__(model, save, verbose)

    def __call__(self, X, **kwargs):

        """__call__(model, **kwargs)

        Evaluate a list X of solutions with the original loss function.

        Parameters
        ----------
        X : list
            List of solutions to evaluate. be carefull if a solution is a list X must be a list of lists.
        **kwargs : dict, optional
            Additionnal informations to save before the score.

        Returns
        -------
        res : list
            Return a list of all the scores corresponding to each evaluated solution of X.

        """

        self.build_bar(len(X))

        self.new_best = False

        res = []

        for x in X:
            outputs, trained_model = self._build_return(self.model(x))
            score, others = outputs["score"], outputs

            res.append(score)

            self.calls += 1

            # Saving
            if self.save:
                self._save_file(x, score, **others, **kwargs)

                if trained_model:
                    self._save_model(score, trained_model)

            self._save_best(x, score)

        self.close_bar()
        return res

    def _save_model(self, score, trained_model):
        # Save model into a file if it is better than the best found one
        if score < self.best_score:
            save_path = os.path.join(
                self.model_path, f"{self.model.__class__.__name__}_best"
            )
            if hasattr(trained_model, "save") and callable(
                getattr(trained_model, "save")
            ):
                os.system(f"rm -rf {save_path}")
                trained_model.save(save_path)
            else:
                logger.error(
                    "Model/loss function does not have a method called `save`"
                )
                exit()


# Wrap different loss functions
def Loss(model=None, save=False, verbose=True, MPI=False):
    """Loss(model=None, save=False, verbose=True, MPI=False)

    Wrap a function of type :math:`f(x)=y`. See `LossFunc` for more info.

    Parameters
    ----------
    model : function, default=None
        Function of type f(x)=y. x must be a solution. A solution can be a list of float, int... It can also be of mixed types, containing, strings, float, int...

    save : string, optional
        Filename where to save the best found model. Only one model is saved for memory issues.

    MPI : boolean, optional
        Wrap the function with MPILoss if True, with SerialLoss else.

    Returns
    -------
    wrapper : LossFunc
        Wrapped original function

    Examples
    --------
    >>> import numpy as np
    >>> from zellij.core.loss_func import Loss
    >>> @Loss(save=False, verbose=True)
    ... def himmelblau(x):
    ...   x_ar = np.array(x)
    ...   return np.sum(x_ar**4 -16*x_ar**2 + 5*x_ar) * (1/len(x_ar))
    >>> print(f"Best solution found: f({himmelblau.best_sol}) = {himmelblau.best_score}")
    Best solution found: f(None) = inf
    >>> print(f"Number of evaluations:{himmelblau.calls}")
    Number of evaluations:0
    >>> print(f"All evaluated solutions:{himmelblau.all_solutions}")
    All evaluated solutions:[]
    >>> print(f"All loss values:{himmelblau.all_scores}")
    All loss values:[]



    """
    if model:
        return SerialLoss(model)
    else:

        def wrapper(model):
            if MPI:
                return MPILoss(model, save, verbose)
            else:
                return SerialLoss(model, save, verbose)

        return wrapper



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:46:31+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:45:59+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


import numpy as np


def shifted_cigar_bent(y, bias=100):
    y = np.array(y)
    shifted = y + 1 / len(y)
    return shifted[0] ** 2 + (10 ** 6) * np.sum(np.square(shifted[1:])) + bias


def shifted_rastrigin(y, bias=200):
    y = np.array(y)
    shifted = y + 1 / len(y)
    return (
        np.sum(
            np.add(
                np.subtract(
                    np.square(shifted), 10 * np.cos(2 * np.pi * shifted)
                ),
                10,
            )
        )
        + bias
    )


def himmelblau(y):
    y = np.array(y)
    return np.sum(y ** 4 - 16 * y ** 2 + 5 * y) * (1 / len(y))


def alpine(y):
    y = np.array(y)
    return 1 / len(y) * np.sum(np.absolute(y * np.sin(y) + 0.1 * y))


def ackley(y, bias=1100):
    y = np.array(y)
    return (
        -20 * np.exp(-0.2 * np.sqrt(1 / len(y) * np.sum(y ** 2)))
        - np.exp(1 / len(y) * np.sum(np.cos(np.pi * 2 * y)))
        + 20
        + np.exp(1)
        + bias
    )


def happycat(y, bias=100):
    y = np.array(y)
    return (
        np.absolute(np.sum(np.square(y) - len(y))) ** (1 / 4)
        + 0.5 / len(y) * (np.sum(np.square(y)) - np.sum(y))
        + 0.5
        + bias
    )


def shifted_levy(y, bias=500):
    y = np.array(y)
    return (
        np.sin(np.pi * y[0]) ** 2
        + np.sum((y[:-1] - 1) ** 2 * (1 + 10 * np.sin(np.pi * y[:-1] + 1) ** 2))
        + (y[-1] - 1) ** 2 * (1 + np.sin(2 * np.pi * y[-1]) ** 2)
        + bias
    )


def brown(y):
    y = np.array(y)
    return np.sum(
        (y[:-1] ** 2) ** (y[1:] ** 2 + 1) + (y[1:] ** 2) ** (y[:-1] ** 2 + 1)
    )


def shifted_rotated_rosenbrock(y, bias=700):
    y = np.array(y)
    return (
        np.sum(
            (
                np.square(100 * np.square(y[1:] - np.square(y[:-1])))
                + np.square(y[:-1] - 1)
            )
        )
        + bias
    )


def random(y):
    y = np.array(y)
    return np.random.randint(1) * np.random.random() * 1000 + 1000



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:44:11+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


import numpy as np
import abc
import copy

import time

from zellij.strategies.utils.spoke_dart import (
    randomMuller,
    Hyperplane,
    HalfLine,
)

import logging

logger = logging.getLogger("zellij.fractal")
logger.setLevel(logging.INFO)


class Fractal(object):
    """Fractal

    Fractal is an abstract class used in Fractal Decomposition. This class is used to build a partition tree of fractals. Each object contains a reference to its father, references to its children, its bounds,\
    and its heuristic value (the score) computed after each exploration. Fractals are simplified and continuous subspaces, builded thanks to the original SearchSpace object.

    Attributes
    ----------

    lo_bounds : list[float]
        Contains the lower bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

    up_bounds : list[float]
        Contains the upper bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

    id : int
        Identifier of a fractal. Combined to the id of itf parents, the id is unique.

    father : Fractal
        Reference to the parent of the current fractal.

    children : list[Fractal]
        References to all children of the current fractal.

    score : {float, int}
        Heuristic value associated to the fractal after an exploration

    level : int
        Current level of the fractal in the partition tree. See Tree_search.

    min_score : {float, int}
        Score associated to the best found solution inside the fractal

    best_sol : list[{float, int, str}]
        Best found solution inside the fractal in its mixed format.

    all_solutions : float
        Historic of all evaluated solutions inside the fractal.

    all_scores : float
        Historic of all evaluated scores inside the fractal..

    Methods
    -------

    add_point(score, solution)
        Adds a point to the fractal

    create_children()
        Abstract method which defines how fractal children should be created


    See Also
    --------
    LossFunc : Defines what a loss function is
    Tree_search : Defines how to explore and exploit a fractal partition tree.
    SearchSpace : Initial search space used to build fractal.
    Hypercube : Inherited Fractal type
    Hypersphere : Inherited Fractal type
    """

    def __init__(
        self,
        lo_bounds,
        up_bounds,
        father="root",
        level=0,
        id=0,
        children=[],
        score=None,
    ):

        """__init__(lo_bounds, up_bounds, father="root", level=0, id=0, children=[], score=None)

        Parameters
        ----------
        lo_bounds : list[float]
            Contains the lower bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        up_bounds : list[float]
            Contains the upper bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        father : Fractal, default='root'
            Reference to the parent of the current fractal.

        level : int, default=0
            Current level of the fractal in the partition tree. See Tree_search.

        id : int, default=0
            Identifier of a fractal. Combined to the id of itf parents, the id is unique.

        children : list[Fractal], default=[]
            References to all children of the current fractal. If no child is given, children will be built by the method create_children during the tree building.

        score : {float, int}, default=None
            Heuristic value associated to the fractal after an exploration.  If no score is given, it will be built after the execution of an exploration strategy inside the fractal.

        """

        self.lo_bounds = np.array(lo_bounds)
        self.up_bounds = np.array(up_bounds)

        self.id = id
        self.father = father
        self.children = []
        self.score = score
        self.level = level

        self.min_score = float("inf")
        self.best_sol = None
        self.solutions = []
        self.all_scores = []

    def add_point(self, score, solution):

        """add_point(self,score, solution)

        This method adds a point associated to its evaluation by the loss function (f(solution)=score), to the historic of the fractal,\
         and determine if this point is the best one among all evaluated points inside the fractal.

        Parameters
        ----------
        score : {int, float}
            Score associated to the evaluated solution

        solution : list[{int, float, str}]
            It corresponds to a point in the mixed format inside the fractal and associated to its evaluation (score).

        """
        for sol, sco in zip(solution, score):
            self.solutions.append(sol)
            self.all_scores.append(sco)

            if sco < self.min_score:
                self.min_score = sco
                self.best_sol = sol

    @abc.abstractmethod
    def create_children(self):
        """create_children(self)

        Abstract method which will create children of the current Fractal object, according to certain rules (Hypercube, Hypersphere...)

        """
        pass


class Hypercube(Fractal):
    """Hypercube

    The hypercube is a basic hypervolume to decompose the SearchSpace. It's also one of the most computationally inefficient in high dimension.\
    The decomposition complexity of an Hypercube with equalsize Hypercubes, is equal to $2^d$, d is the dimension.\
    However building a single hypercube is low complexity task, and the space coverage is very good, 100% of the initial hypercube is covered by its children.

    Attributes
    ----------

    dim : int
        Number of dimensions

    Methods
    -------

    create_children(self)
        Method which defines how to build children Hypercubes based on the current Hypercube.


    See Also
    --------
    LossFunc : Defines what a loss function is
    Tree_search : Defines how to explore and exploit a fractal partition tree.
    SearchSpace : Initial search space used to build fractal.
    Fractal : Parent class. Basic object to define what a fractal is.
    Hypersphere : Another hypervolume, with different properties
    """

    def __init__(
        self,
        lo_bounds,
        up_bounds,
        father="root",
        level=0,
        id=0,
        children=[],
        score=None,
    ):

        """__init__(lo_bounds, up_bounds, father="root", level=0, id=0, children=[], score=None)

        Parameters
        ----------
        lo_bounds : list[float]
            Contains the lower bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        up_bounds : list[float]
            Contains the upper bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        father : Fractal, default='root'
            Reference to the parent of the current fractal.

        level : int, default=0
            Current level of the fractal in the partition tree. See Tree_search.

        id : int, default=0
            Identifier of a fractal. Combined to the id of itf parents, the id is unique.

        children : list[Fractal], default=[]
            References to all children of the current fractal. If no child is given, children will be built by the method create_children during the tree building.

        score : {float, int}, default=None
            Heuristic value associated to the fractal after an exploration.  If no score is given, it will be built after the execution of an exploration strategy inside the fractal.

        """

        super().__init__(
            lo_bounds, up_bounds, father, level, id, children, score
        )

        self.dim = len(self.up_bounds)

    def create_children(self):

        """create_children(self)

        Method which defines how to build children Hypercubes based on the current Hypercube.
        It uses Hyperplan bisecting to build children. To build an Hypercube, it only requires lower and upper bounds.

        """

        level = self.level + 1

        up_m_lo = self.up_bounds - self.lo_bounds
        radius = np.abs(up_m_lo / 2)
        bounds = [[self.lo_bounds, self.up_bounds]]

        # Hyperplan bisecting
        next_b = []
        for i in range(self.dim):
            next_b = []
            for b in bounds:

                # First part
                up = np.copy(b[1])
                up[i] = b[0][i] + radius[i]
                next_b.append([np.copy(b[0]), np.copy(up)])

                # Second part
                low = np.copy(b[0])
                low[i] = b[1][i] - radius[i]
                next_b.append([np.copy(low), np.copy(b[1])])

            bounds = copy.deepcopy(next_b)

        # Create Hypercube
        n_h = 0
        for b in bounds:
            h = Hypercube(b[0], b[1], self, level, n_h)
            self.children.append(h)
            n_h += 1

    def __repr__(self):
        if type(self.father) == str:
            id = "root"
        else:
            id = str(self.father.id)

        return (
            "ID: "
            + str(self.id)
            + " son of "
            + id
            + "\n"
            + "BOUNDS: "
            + str(self.lo_bounds)
            + "|"
            + str(self.up_bounds)
            + "\n"
        )


class Hypersphere2(Fractal):

    """Hypersphere

    The Hypersphere is a basic hypervolume to decompose the SearchSpace. It is one of the most computationally efficient, to decompose the SearchSpace.
    To decompose an hypersphere by equalsize hypersphere the complexity is equal to $2*d$, d is the dimension, moreover building an hypersphere is easy, it only needs a center and its radius.
    However the space coverage is poor, indeed the volume of an hypersphere tends to 0, when the dimension tends to infinity. To partially tackle this problem, an inflation rate allows bigger hypersphere,
    but this will create overlapping hypervolumes.

    Attributes
    ----------

    dim : int
        Number of dimensions

    inflation : float
        Inflation rate of hyperspheres. Be carefull a too large inflation can result to hypersphere with identical center and radius.

    center : list[float]
        List of floats containing the coordinates

    radius : list[float]
        List of floats containing the radius for each dimension (in case the initial SearchSpace is not an hypercube).

    Methods
    -------

    create_children(self)
        Method which defines how to build children Hyperspheres based on the current Hypersphere.


    See Also
    --------
    LossFunc : Defines what a loss function is
    Tree_search : Defines how to explore and exploit a fractal partition tree.
    SearchSpace : Initial search space used to build fractal.
    Fractal : Parent class. Basic object to define what a fractal is.
    Hypercube : Another hypervolume, with different properties
    """

    def __init__(
        self,
        lo_bounds,
        up_bounds,
        father="root",
        level=0,
        id=0,
        inflation=1.75,
        children=[],
        score=None,
    ):

        """__init__(lo_bounds, up_bounds, father="root", level=0, id=0, inflation=1.75, children=[], score=None)

        Parameters
        ----------
        lo_bounds : list[float]
            Contains the lower bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        up_bounds : list[float]
            Contains the upper bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        father : Fractal, default='root'
            Reference to the parent of the current fractal.

        level : int, default=0
            Current level of the fractal in the partition tree. See Tree_search.

        id : int, default=0
            Identifier of a fractal. Combined to the id of itf parents, the id is unique.

        children : list[Fractal], default=[]
            References to all children of the current fractal.
            If no child IS given, children will be built by the method create_children during the tree building.

        score : {float, int}, default=None
            Heuristic value associated to the fractal after an exploration.  If no score is given, it will be built after the execution of an exploration strategy inside the fractal.

        """

        super().__init__(
            lo_bounds, up_bounds, father, level, id, children, score
        )

        self.dim = len(self.up_bounds)

        up_m_lo = self.up_bounds - self.lo_bounds
        center = self.lo_bounds + (up_m_lo) / 2
        radius = up_m_lo[0] / 2

        self.center = center
        self.radius = radius

        self.inflation = inflation

        self.radius = self.radius * self.inflation

    def create_children(self):

        """create_children(self)

        Method which defines how to build children Hypercubes based on the current Hypercube.
        It uses Hyperplan bisecting to build children. To build an Hypercube, it only requires lower and upper bounds.

        """

        level = self.level + 1

        r_p = self.radius / (1 + np.sqrt(2))

        n_h = 0
        for i in range(self.dim):

            n_h += 1
            center = np.copy(self.center)
            center[i] += ((-1) ** i) * (self.radius - r_p)

            lo = np.maximum(center - r_p, self.lo_bounds)
            up = np.minimum(center + r_p, self.up_bounds)

            h = Hypersphere(lo, up, self, level, n_h, inflation=self.inflation)
            self.children.append(h)

            n_h += 1
            center = np.copy(self.center)
            center[i] -= ((-1) ** i) * (self.radius - r_p)

            lo = np.maximum(center - r_p, self.lo_bounds)
            up = np.minimum(center + r_p, self.up_bounds)

            h = Hypersphere(lo, up, self, level, n_h, inflation=self.inflation)
            self.children.append(h)

    def __repr__(self):
        if type(self.father) == str:
            id = "root"
        else:
            id = str(self.father.id)

        return (
            "ID: "
            + str(self.id)
            + " son of "
            + id
            + " at level "
            + str(self.level)
            + "\n"
            + "BOUNDS: "
            + str(self.lo_bounds)
            + "|"
            + str(self.up_bounds)
            + "\n"
        )


class Hypersphere(Fractal):

    """Hypersphere

    The Hypersphere is a basic hypervolume to decompose the SearchSpace. It is one of the most computationally efficient, to decompose the SearchSpace.
    To decompose an hypersphere by equalsize hypersphere the complexity is equal to $2*d$, d is the dimension, moreover building an hypersphere is easy, it only needs a center and its radius.
    However the space coverage is poor, indeed the volume of an hypersphere tends to 0, when the dimension tends to infinity. To partially tackle this problem, an inflation rate allows bigger hypersphere,
    but this will create overlapping hypervolumes.

    Attributes
    ----------

    dim : int
        Number of dimensions

    inflation : float
        Inflation rate of hyperspheres. Be carefull a too large inflation can result to hypersphere with identical center and radius.

    center : list[float]
        List of floats containing the coordinates

    radius : list[float]
        List of floats containing the radius for each dimension (in case the initial SearchSpace is not an hypercube).

    Methods
    -------

    create_children(self)
        Method which defines how to build children Hyperspheres based on the current Hypersphere.


    See Also
    --------
    LossFunc : Defines what a loss function is
    Tree_search : Defines how to explore and exploit a fractal partition tree.
    SearchSpace : Initial search space used to build fractal.
    Fractal : Parent class. Basic object to define what a fractal is.
    Hypercube : Another hypervolume, with different properties
    """

    def __init__(
        self,
        lo_bounds,
        up_bounds,
        father="root",
        level=0,
        id=0,
        radius=None,
        center=None,
        inflation=1.75,
        children=[],
        score=None,
    ):

        """__init__(father,lo_bounds,up_bounds,level,id,children=[],score=None)

        Parameters
        ----------
        lo_bounds : list[float]
            Contains the lower bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.
            /!\ Those bounds are clipped in case the hypersphere is outside the decision space

        up_bounds : list[float]
            Contains the upper bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.
            /!\ Those bounds are clipped in case the hypersphere is outside the decision space

        father : Fractal, default='root'
            Reference to the parent of the current fractal.

        level : int, default=0
            Current level of the fractal in the partition tree. See Tree_search.

        id : int, default=0
            Identifier of a fractal. Combined to the id of its parents, the id is unique.

        children : list[Fractal], default=[]
            References to all children of the current fractal.
            If no child is given, children will be built by the method create_children during the tree building.

        score : {float, int}, default=None
            Heuristic value associated to the fractal after an exploration.
            If no score is given, it will be built after the execution of an exploration inside the fractal.

        """

        super().__init__(
            lo_bounds, up_bounds, father, level, id, children, score
        )

        if isinstance(father, str):
            self.original_up = self.lo_bounds
            self.original_down = self.up_bounds
        else:
            self.original_up = father.original_up
            self.original_down = father.original_down

        self.dim = len(self.up_bounds)

        if center is None or radius is None:
            up_m_lo = self.up_bounds - self.lo_bounds
            center = self.lo_bounds + (up_m_lo) / 2
            radius = up_m_lo[0] / 2

        self.center = center
        self.inflation = inflation
        self.radius = radius * self.inflation

    def create_children(self):

        """create_children(self)

        Method which defines how to build children Hypercubes based on the current Hypercube.
        It uses Hyperplan bisecting to build children. To build an Hypercube, it only requires lower and upper bounds.

        """

        level = self.level + 1

        r_p = self.radius / (1 + np.sqrt(2))

        n_h = 0
        for i in range(self.dim):

            logger.info(f"Building children n°{n_h}/{self.dim*2}")

            center = np.copy(self.center)
            center[i] += r_p
            center[i] = np.minimum(center[i], 1)

            lo = center - r_p
            lo[lo < 0] = 0
            lo[lo > 1] = 1

            up = center + r_p
            up[up < 0] = 0
            up[up > 1] = 1

            h = Hypersphere(
                lo,
                up,
                self,
                level,
                n_h,
                center=np.copy(center),
                radius=r_p,
                inflation=self.inflation,
            )

            self.children.append(h)
            n_h += 1

            logger.info(f"Building children n°{n_h}/{self.dim*2}")

            center = np.copy(self.center)
            center[i] -= r_p
            center[i] = np.maximum(center[i], 0)

            lo = center - r_p
            lo[lo < 0] = 0
            lo[lo > 1] = 1

            up = center + r_p
            up[up < 0] = 0
            up[up > 1] = 1

            h = Hypersphere(
                lo,
                up,
                self,
                level,
                n_h,
                center=np.copy(center),
                radius=r_p,
                inflation=self.inflation,
            )

            self.children.append(h)
            n_h += 1

    def __repr__(self):
        if type(self.father) == str:
            id = "root"
        else:
            id = str(self.father.id)

        return (
            "ID: "
            + str(self.id)
            + " son of "
            + id
            + " at level "
            + str(self.level)
            + "\n"
            + "BOUNDS: "
            + str(self.lo_bounds)
            + "|"
            + str(self.up_bounds)
            + "\n"
        )


class Section(Fractal):

    """Section

    Performs a n-Section of the search space. 3-Section: DIRECT, 2-Section: BIRECT

    Attributes
    ----------

    dim : int
        Number of dimensions

    Methods
    -------

    create_children(self)
        Method which defines how to build children Hyper-rectangles based on the current Hyper-rectangle.


    See Also
    --------
    LossFunc : Defines what a loss function is
    Tree_search : Defines how to explore and exploit a fractal partition tree.
    SearchSpace : Initial search space used to build fractal.
    Fractal : Parent class. Basic object to define what a fractal is.
    Hypercube : Another hypervolume, with different properties
    """

    def __init__(
        self,
        lo_bounds,
        up_bounds,
        father="root",
        level=0,
        id=0,
        children=[],
        score=None,
        n=2,
    ):

        """__init__(lo_bounds, up_bounds, father="root", level=0, id=0, children=[], score=None, n=2)

        Parameters
        ----------
        lo_bounds : list[float]
            Contains the lower bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        up_bounds : list[float]
            Contains the upper bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        father : Fractal, default='root'
            Reference to the parent of the current fractal.

        level : int, default=0
            Current level of the fractal in the partition tree. See Tree_search.

        id : int, default=0
            Identifier of a fractal. Combined to the id of itf parents, the id is unique.

        children : list[Fractal], default=[]
            References to all children of the current fractal. If no child is given, children will be built by the method create_children during the tree building.

        score : {float, int}, default=None
            Heuristic value associated to the fractal after an exploration.  If no score is given, it will be built after the execution of an exploration strategy inside the fractal.

        """

        super().__init__(
            lo_bounds, up_bounds, father, level, id, children, score
        )

        assert n > 1, logger.error(
            f"{n}-Section is not possible, n must be > 1"
        )

        self.dim = len(self.up_bounds)
        self.section = n

        up_m_lo = self.up_bounds - self.lo_bounds
        self.center = self.lo_bounds + (up_m_lo) / 2
        self.longest = np.argmax(up_m_lo)
        self.length = up_m_lo[self.longest]

    def create_children(self):

        level = self.level + 1

        n_h = 0

        new_val = self.length / self.section

        lo = np.copy(self.lo_bounds)
        up = np.copy(self.up_bounds)
        up[self.longest] = lo[self.longest] + new_val

        for i in range(self.section):

            h = Section(lo, up, self, level, n_h, n=self.section)

            self.children.append(h)

            lo = np.copy(h.lo_bounds)
            up = np.copy(h.up_bounds)
            lo[self.longest] += new_val
            up[self.longest] += new_val

            n_h += 1

    def __repr__(self):
        if type(self.father) == str:
            id = "root"
        else:
            id = str(self.father.id)

        return (
            "ID: "
            + str(self.id)
            + " son of "
            + id
            + " at level "
            + str(self.level)
            + "\n"
            + "BOUNDS: "
            + str(self.lo_bounds)
            + "|"
            + str(self.up_bounds)
            + "\n"
        )


class Voronoi(Fractal):
    def __init__(
        self,
        lo_bounds,
        up_bounds,
        father="root",
        level=0,
        id=0,
        children=[],
        score=None,
        seed="random",
        n_seeds=None,
    ):

        """lo_bounds, up_bounds, father="root", level=0, id=0, children=[], score=None, seed="random", spokes=2, n_seeds=5

        Parameters
        ----------
        lo_bounds : list[float]
            Contains the lower bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        up_bounds : list[float]
            Contains the upper bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        father : Fractal, default='root'
            Reference to the parent of the current fractal.

        level : int, default=0
            Current level of the fractal in the partition tree. See Tree_search.

        id : int, default=0
            Identifier of a fractal. Combined to the id of itf parents, the id is unique.

        children : list[Fractal], default=[]
            References to all children of the current fractal. If no child is given, children will be built by the method create_children during the tree building.

        score : {float, int}, default=None
            Heuristic value associated to the fractal after an exploration.  If no score is given, it will be built after the execution of an exploration strategy inside the fractal.

        seed : {random, [[float]]}
            If 'random' Voronoi centroid will be initialized randomly. Else: if a list of points is given, the DynamicVoronoi will use them.

        spokes : int, default=2
            Number of randoms spokes to draw during hyperplane sampling (SpokeDart)

        n_seeds : int, default=2
            Number of centroids at each decomposition.

        """

        super().__init__(
            lo_bounds, up_bounds, father, level, id, children, score
        )

        self.dim = len(self.up_bounds)
        if n_seeds is None:
            self.n_seeds = 2 * self.dim
        else:
            self.n_seeds = n_seeds

        self.next_seeds = []

        if isinstance(seed, str) and seed == "random":
            self.all_seeds = []
            self.next_seeds = list(
                np.random.random((self.n_seeds, self.dim))
                * (np.array(self.up_bounds) - np.array(self.lo_bounds))
                + np.array(self.lo_bounds)
            )
            self.seed = "root"
            self.hyperplanes = []

        else:
            self.all_seeds = self.father.all_seeds
            self.hyperplanes = []
            self.seed = seed
            self.center = self.seed

        self.xlist = np.zeros(2 * self.dim)

    @abc.abstractmethod
    def create_children(self):
        pass

    @abc.abstractmethod
    def update(self):
        pass

    def randomSpoke(self, s):
        p = randomMuller(1, self.dim)[0]
        pfar = p + s

        return HalfLine(s, pfar)

    def dimSpoke(self, s, dim, r=1):
        p = np.zeros(self.dim)
        p[dim] = r
        pfar = p + s
        return HalfLine(s, pfar)

    def shiftBorder(self, s, p):
        l = HalfLine(s, p)
        return l.point(np.random.random())

    def clipBorder(self, line, upma, loma):

        self.xlist[: self.dim] = loma / (line.v * (1 + 1e-10))
        self.xlist[self.dim :] = upma / (line.v * (1 + 1e-10))

        x = np.nanmin(np.where(self.xlist > 0, self.xlist, np.inf))
        res = line.point(x)

        return res


class DynamicVoronoi(Voronoi):
    def __init__(
        self,
        lo_bounds,
        up_bounds,
        father="root",
        level=0,
        id=0,
        children=[],
        score=None,
        seed="random",
        spokes=2,
        n_seeds=5,
    ):

        """lo_bounds, up_bounds, father="root", level=0, id=0, children=[], score=None, seed="random", spokes=2, n_seeds=5

        Parameters
        ----------
        lo_bounds : list[float]
            Contains the lower bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        up_bounds : list[float]
            Contains the upper bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        father : Fractal, default='root'
            Reference to the parent of the current fractal.

        level : int, default=0
            Current level of the fractal in the partition tree. See Tree_search.

        id : int, default=0
            Identifier of a fractal. Combined to the id of itf parents, the id is unique.

        children : list[Fractal], default=[]
            References to all children of the current fractal. If no child is given, children will be built by the method create_children during the tree building.

        score : {float, int}, default=None
            Heuristic value associated to the fractal after an exploration.  If no score is given, it will be built after the execution of an exploration strategy inside the fractal.

        seed : {random, [[float]]}
            If 'random' Voronoi centroid will be initialized randomly. Else: if a list of points is given, the DynamicVoronoi will use them.

        spokes : int, default=2
            Number of randoms spokes to draw during hyperplane sampling (SpokeDart)

        n_seeds : int, default=2
            Number of centroids at each decomposition.

        """

        super().__init__(
            lo_bounds,
            up_bounds,
            father,
            level,
            id,
            children,
            score,
            seed=seed,
            n_seeds=n_seeds,
        )

        self.n_dim = 2 * self.dim
        self.spokes = spokes

        self.sampled_bounds = []

    def create_children(self):

        update_neighbors = False

        if not isinstance(self.father, str):
            # Current cell will be it's own children
            selfchild = DynamicVoronoi(
                self.lo_bounds,
                self.up_bounds,
                self,
                self.level + 1,
                self.id,
                seed=self.seed,
                spokes=self.spokes,
                n_seeds=self.n_seeds,
            )
            self.children.append(selfchild)
            # Replace previous cell by new cell
            self.all_seeds[self.id] = selfchild

        for i in self.next_seeds:
            child = DynamicVoronoi(
                self.lo_bounds,
                self.up_bounds,
                self,
                self.level + 1,
                len(self.all_seeds),
                seed=i,
                spokes=self.spokes,
                n_seeds=self.n_seeds,
            )
            self.children.append(child)
            self.all_seeds.append(child)

        for child in self.children:
            for cell in self.all_seeds:
                cell.sampled_bounds = []
                cell.next_seeds = []
                if child.id != cell.id:
                    try:
                        h = Hyperplane(self.children[i], self.children[j])
                        self.children[i].hyperplanes.append(h)
                        self.children[j].hyperplanes.append(h)
                    except AssertionError as e:
                        logger.warning(f"Hyperplane building aborted: {e}")

        for i, c in enumerate(self.all_seeds):
            logger.info(f"Building children n°{i}/{len(self.children)}")
            if len(c.hyperplanes) > 0:
                c.update()
            else:
                self.children.pop(i)

    def update(self):
        upma = self.up_bounds - self.seed
        loma = self.lo_bounds - self.seed

        cell_idx = [False] * len(self.hyperplanes)

        # Fixed points (2 for each dimension)
        for d in range(self.n_dim):

            l = self.dimSpoke(self.seed, d % self.dim)

            inter = np.empty((0, self.dim))

            for j, h in enumerate(self.hyperplanes):

                on, pfar = h.intersection(l)

                if on:
                    # Clip line to bounds
                    if np.any(pfar > self.up_bounds) or np.any(
                        pfar < self.lo_bounds
                    ):

                        pfar_clipped = self.clipBorder(l, upma, loma)
                        inter = np.append(inter, [pfar_clipped], axis=0)

                        cell_idx[j] = False

                    else:
                        inter = np.append(inter, [pfar], axis=0)

                        cell_idx[j] = True
                else:
                    cell_idx[j] = False

            if len(inter) == 0:

                pfar_clipped = self.clipBorder(l, upma, loma)
                self.sampled_bounds.append(np.copy(pfar_clipped))

            else:
                dist = np.linalg.norm(inter - l.A, axis=1)
                minidx = np.argmin(dist)

                a = np.copy(inter[minidx])

                if cell_idx[minidx]:
                    self.sampled_hyperplanes.add(minidx)

                    if self.hyperplanes[minidx].cellX != self:
                        self.hyperplanes[minidx].cellX.sampled_bounds.append(a)

                    else:
                        self.hyperplanes[minidx].cellY.sampled_bounds.append(a)

                self.sampled_bounds.append(a)

        # Random points
        for d in range(self.spokes):

            l = self.randomSpoke(self.seed)

            inter = np.empty((0, self.dim))

            for j, h in enumerate(self.hyperplanes):

                on, pfar = h.intersection(l)

                if on:
                    # Clip line to bounds
                    if np.any(pfar > self.up_bounds) or np.any(
                        pfar < self.lo_bounds
                    ):

                        pfar_clipped = self.clipBorder(l, upma, loma)
                        inter = np.append(inter, [pfar_clipped], axis=0)

                        cell_idx[j] = False

                    else:
                        inter = np.append(inter, [pfar], axis=0)

                        cell_idx[j] = True
                else:
                    cell_idx[j] = False

            if len(inter) == 0:

                pfar_clipped = self.clipBorder(l, upma, loma)
                self.sampled_bounds.append(np.copy(pfar_clipped))

            else:
                dist = np.linalg.norm(inter - l.A, axis=1)
                minidx = np.argmin(dist)

                a = np.copy(inter[minidx])

                if cell_idx[minidx]:
                    self.sampled_hyperplanes.add(minidx)

                    if self.hyperplanes[minidx].cellX != self:
                        self.hyperplanes[minidx].cellX.sampled_bounds.append(a)

                    else:
                        self.hyperplanes[minidx].cellY.sampled_bounds.append(a)

                self.sampled_bounds.append(a)

        dist = np.linalg.norm(np.array(self.sampled_bounds) - self.seed, axis=1)

        dist = np.nan_to_num(dist)
        sum = np.sum(dist)

        if sum != 0:
            p = dist / sum
            choosen = np.random.choice(
                list(range(len(self.sampled_bounds))),
                np.minimum(np.count_nonzero(p), self.n_seeds),
                replace=False,
                p=p,
            )

            for c in choosen:
                self.next_seeds.append(
                    self.shiftBorder(self.seed, self.sampled_bounds[c])
                )


class FixedVoronoi(Voronoi):
    def __init__(
        self,
        lo_bounds,
        up_bounds,
        father="root",
        level=0,
        id=0,
        children=[],
        score=None,
        seed="random",
        spokes=2,
        n_seeds=5,
    ):

        """__init__(lo_bounds, up_bounds, father="root", level=0, id=0, children=[], score=None, seed="random", spokes=2, n_seeds=5)

        Parameters
        ----------
        lo_bounds : list[float]
            Contains the lower bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        up_bounds : list[float]
            Contains the upper bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        father : Fractal, default='root'
            Reference to the parent of the current fractal.

        level : int, default=0
            Current level of the fractal in the partition tree. See Tree_search.

        id : int, default=0
            Identifier of a fractal. Combined to the id of itf parents, the id is unique.

        children : list[Fractal], default=[]
            References to all children of the current fractal. If no child is given, children will be built by the method create_children during the tree building.

        score : {float, int}, default=None
            Heuristic value associated to the fractal after an exploration.  If no score is given, it will be built after the execution of an exploration strategy inside the fractal.

        seed : {random, [[float]]}
            If 'random' Voronoi centroid will be initialized randomly. Else: if a list of points is given, the DynamicVoronoi will use them.

        spokes : int, default=2
            Number of randoms spokes to draw during hyperplane sampling (SpokeDart)

        n_seeds : int, default=2
            Number of centroids at each decomposition.

        """

        super().__init__(
            lo_bounds,
            up_bounds,
            father,
            level,
            id,
            children,
            score,
            seed=seed,
            n_seeds=n_seeds,
        )

        self.dim = len(self.up_bounds)
        self.n_dim = 2 * self.dim
        self.spokes = spokes

        self.sampled_bounds = []

    def create_children(self):

        logger.info(f"Creating children of n°{self.id}")

        # if not isinstance(self.father, str):
        #     # Current cell will be it's own children
        #     selfchild = FixedVoronoi(self.lo_bounds, self.up_bounds, self, self.level + 1, self.id, seed=self.seed, spokes=self.spokes, n_seeds=self.n_seeds)
        #     selfchild.hyperplanes = self.hyperplanes[:]
        #     self.children.append(selfchild)
        #
        #     # Replace previous cell by new cell
        #     self.all_seeds.append(selfchild)

        for i in self.next_seeds:
            child = FixedVoronoi(
                self.lo_bounds,
                self.up_bounds,
                self,
                self.level + 1,
                len(self.all_seeds),
                seed=i,
                spokes=self.spokes,
                n_seeds=self.n_seeds,
            )
            child.hyperplanes = self.hyperplanes[:]
            self.children.append(child)
            self.all_seeds.append(child)

        for i in range(len(self.children) - 1):
            for j in range(i + 1, len(self.children)):
                try:
                    h = Hyperplane(self.children[i], self.children[j])
                    self.children[i].hyperplanes.append(h)
                    self.children[j].hyperplanes.append(h)
                except AssertionError as e:
                    logger.warning(f"Hyperplane building aborted: {e}")

        for i, c in enumerate(self.children):
            logger.info(f"Building children n°{i}/{len(self.children)}")
            if len(c.hyperplanes) > 0:
                c.update()
            else:
                self.children.pop(i)

    def update(self):

        upma = self.up_bounds - self.seed
        loma = self.lo_bounds - self.seed

        cell_idx = [False] * len(self.hyperplanes)

        # Fixed points (2 for each dimension)
        for d in range(self.n_dim):

            l = self.dimSpoke(self.seed, d % self.dim)

            inter = np.empty((0, self.dim))

            for j, h in enumerate(self.hyperplanes):

                on, pfar = h.intersection(l)

                if on:
                    # Clip line to bounds
                    if np.any(pfar > self.up_bounds) or np.any(
                        pfar < self.lo_bounds
                    ):

                        pfar_clipped = self.clipBorder(l, upma, loma)
                        inter = np.append(inter, [pfar_clipped], axis=0)

                        cell_idx[j] = False

                    else:
                        inter = np.append(inter, [pfar], axis=0)

                        cell_idx[j] = True
                else:
                    cell_idx[j] = False

            if len(inter) == 0:

                pfar_clipped = self.clipBorder(l, upma, loma)
                self.sampled_bounds.append(np.copy(pfar_clipped))

            else:
                dist = np.linalg.norm(inter - l.A, axis=1)
                minidx = np.argmin(dist)

                a = np.copy(inter[minidx])

                if cell_idx[minidx]:

                    if self.hyperplanes[minidx].cellX != self:
                        self.hyperplanes[minidx].cellX.sampled_bounds.append(a)

                    else:
                        self.hyperplanes[minidx].cellY.sampled_bounds.append(a)

                self.sampled_bounds.append(a)

        # Random points
        for d in range(self.spokes):

            l = self.randomSpoke(self.seed)

            inter = np.empty((0, self.dim))

            for j, h in enumerate(self.hyperplanes):

                on, pfar = h.intersection(l)

                if on:
                    # Clip line to bounds
                    if np.any(pfar > self.up_bounds) or np.any(
                        pfar < self.lo_bounds
                    ):

                        pfar_clipped = self.clipBorder(l, upma, loma)
                        inter = np.append(inter, [pfar_clipped], axis=0)

                        cell_idx[j] = False

                    else:
                        inter = np.append(inter, [pfar], axis=0)

                        cell_idx[j] = True
                else:
                    cell_idx[j] = False

            if len(inter) == 0:

                pfar_clipped = self.clipBorder(l, upma, loma)
                self.sampled_bounds.append(np.copy(pfar_clipped))

            else:
                dist = np.linalg.norm(inter - l.A, axis=1)
                minidx = np.argmin(dist)

                a = np.copy(inter[minidx])

                if cell_idx[minidx]:

                    if self.hyperplanes[minidx].cellX != self:
                        self.hyperplanes[minidx].cellX.sampled_bounds.append(a)

                    else:
                        self.hyperplanes[minidx].cellY.sampled_bounds.append(a)

                self.sampled_bounds.append(a)

        dist = np.linalg.norm(np.array(self.sampled_bounds) - self.seed, axis=1)

        dist = np.nan_to_num(dist)
        sum = np.sum(dist)

        if sum != 0:
            p = dist / sum
            choosen = np.random.choice(
                list(range(len(self.sampled_bounds))),
                np.minimum(np.count_nonzero(p), self.n_seeds),
                replace=False,
                p=p,
            )

            for c in choosen:
                self.next_seeds.append(
                    self.shiftBorder(self.seed, self.sampled_bounds[c])
                )


class LightFixedVoronoi(Voronoi):
    def __init__(
        self,
        lo_bounds,
        up_bounds,
        father="root",
        level=0,
        id=0,
        children=[],
        score=None,
        seed="random",
        spokes=2,
        n_seeds=5,
    ):

        """lo_bounds, up_bounds, father="root", level=0, id=0, children=[], score=None, seed="random", spokes=2, n_seeds=5

        Parameters
        ----------
        lo_bounds : list[float]
            Contains the lower bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        up_bounds : list[float]
            Contains the upper bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        father : Fractal, default='root'
            Reference to the parent of the current fractal.

        level : int, default=0
            Current level of the fractal in the partition tree. See Tree_search.

        id : int, default=0
            Identifier of a fractal. Combined to the id of itf parents, the id is unique.

        children : list[Fractal], default=[]
            References to all children of the current fractal. If no child is given, children will be built by the method create_children during the tree building.

        score : {float, int}, default=None
            Heuristic value associated to the fractal after an exploration.  If no score is given, it will be built after the execution of an exploration strategy inside the fractal.

        seed : {random, [[float]]}
            If 'random' Voronoi centroid will be initialized randomly. Else: if a list of points is given, the DynamicVoronoi will use them.

        spokes : int, default=2
            Number of randoms spokes to draw during hyperplane sampling (SpokeDart)

        n_seeds : int, default=2
            Number of centroids at each decomposition.

        """

        super().__init__(
            lo_bounds,
            up_bounds,
            father,
            level,
            id,
            children,
            score,
            seed=seed,
            n_seeds=n_seeds,
        )

        self.dim = len(self.up_bounds)
        self.n_dim = 2 * self.dim
        self.spokes = spokes

        self.sampled_bounds = []

    def create_children(self):

        logger.info(f"Creating children of n°{self.id}")

        # if not isinstance(self.father, str):
        #     # Current cell will be it's own children
        #     selfchild = LightFixedVoronoi(self.lo_bounds, self.up_bounds, self, self.level + 1, self.id, seed=self.seed, spokes=self.spokes, n_seeds=self.n_seeds)
        #     selfchild.hyperplanes = self.hyperplanes[:]
        #     self.children.append(selfchild)

        for i in self.next_seeds:
            child = LightFixedVoronoi(
                self.lo_bounds,
                self.up_bounds,
                self,
                self.level + 1,
                len(self.children),
                seed=i,
                spokes=self.spokes,
                n_seeds=self.n_seeds,
            )
            child.hyperplanes = self.hyperplanes[:]

            self.children.append(child)

        for i in range(len(self.children) - 1):
            for j in range(i + 1, len(self.children)):
                try:
                    h = Hyperplane(self.children[i], self.children[j])
                    self.children[i].hyperplanes.append(h)
                    self.children[j].hyperplanes.append(h)
                except AssertionError as e:
                    logger.warning(f"Hyperplane building aborted: {e}")

        for i, c in enumerate(self.children):
            logger.info(f"Building children n°{i}/{len(self.children)}")
            if len(c.hyperplanes) > 0:
                c.update()
            else:
                self.children.pop(i)

    def update(self):

        sampled_hyperplanes = set()

        upma = self.up_bounds - self.seed
        loma = self.lo_bounds - self.seed

        cell_idx = [False] * len(self.hyperplanes)

        # Fixed points (2 for each dimension)
        for d in range(self.n_dim):

            l = self.dimSpoke(self.seed, d % self.dim)

            inter = np.empty((0, self.dim))

            for j, h in enumerate(self.hyperplanes):

                on, pfar = h.intersection(l)

                if on:
                    # Clip line to bounds
                    if np.any(pfar > self.up_bounds) or np.any(
                        pfar < self.lo_bounds
                    ):

                        pfar_clipped = self.clipBorder(l, upma, loma)
                        inter = np.append(inter, [pfar_clipped], axis=0)
                        cell_idx[j] = False
                    else:
                        inter = np.append(inter, [pfar], axis=0)

                        cell_idx[j] = True
                else:
                    cell_idx[j] = False

            if len(inter) == 0:
                pfar_clipped = self.clipBorder(l, upma, loma)
                self.sampled_bounds.append(np.copy(pfar_clipped))

            else:
                dist = np.linalg.norm(inter - l.A, axis=1)
                minidx = np.argmin(dist)

                a = np.copy(inter[minidx])

                if cell_idx[minidx]:
                    sampled_hyperplanes.add(minidx)

                self.sampled_bounds.append(a)

        # Random points
        for d in range(self.spokes):

            l = self.randomSpoke(self.seed)

            inter = np.empty((0, self.dim))

            for j, h in enumerate(self.hyperplanes):

                on, pfar = h.intersection(l)

                if on:
                    # Clip line to bounds
                    if np.any(pfar > self.up_bounds) or np.any(
                        pfar < self.lo_bounds
                    ):

                        pfar_clipped = self.clipBorder(l, upma, loma)
                        inter = np.append(inter, [pfar_clipped], axis=0)
                        cell_idx[j] = False
                    else:
                        inter = np.append(inter, [pfar], axis=0)

                        cell_idx[j] = True
                else:
                    cell_idx[j] = False

            if len(inter) == 0:
                pfar_clipped = self.clipBorder(l, upma, loma)
                self.sampled_bounds.append(np.copy(pfar_clipped))

            else:
                dist = np.linalg.norm(inter - l.A, axis=1)
                minidx = np.argmin(dist)

                a = np.copy(inter[minidx])

                if cell_idx[minidx]:
                    sampled_hyperplanes.add(minidx)

                self.sampled_bounds.append(a)

        self.hyperplanes = [
            self.hyperplanes[idx] for idx in sampled_hyperplanes
        ]

        dist = np.linalg.norm(np.array(self.sampled_bounds) - self.seed, axis=1)

        dist = np.nan_to_num(dist)
        sum = np.sum(dist)

        if sum != 0:
            p = dist / sum
            choosen = np.random.choice(
                list(range(len(self.sampled_bounds))),
                np.minimum(np.count_nonzero(p), self.n_seeds),
                replace=False,
                p=p,
            )

            for c in choosen:
                self.next_seeds.append(
                    self.shiftBorder(self.seed, self.sampled_bounds[c])
                )


class BoxedVoronoi(Voronoi):
    def __init__(
        self,
        lo_bounds,
        up_bounds,
        father="root",
        level=0,
        id=0,
        children=[],
        score=None,
        seed="random",
        spokes=2,
        n_seeds=5,
    ):

        """lo_bounds, up_bounds, father="root", level=0, id=0, children=[], score=None, seed="random", spokes=2, n_seeds=5

        Parameters
        ----------
        lo_bounds : list[float]
            Contains the lower bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        up_bounds : list[float]
            Contains the upper bounds for each dimension of the fractal. Each fractal is bounded by its circumscribed hypercube.

        father : Fractal, default='root'
            Reference to the parent of the current fractal.

        level : int, default=0
            Current level of the fractal in the partition tree. See Tree_search.

        id : int, default=0
            Identifier of a fractal. Combined to the id of itf parents, the id is unique.

        children : list[Fractal], default=[]
            References to all children of the current fractal. If no child is given, children will be built by the method create_children during the tree building.

        score : {float, int}, default=None
            Heuristic value associated to the fractal after an exploration.  If no score is given, it will be built after the execution of an exploration strategy inside the fractal.

        seed : {random, [[float]]}
            If 'random' Voronoi centroid will be initialized randomly. Else: if a list of points is given, the DynamicVoronoi will use them.

        spokes : int, default=2
            Number of randoms spokes to draw during hyperplane sampling (SpokeDart)

        n_seeds : int, default=2
            Number of centroids at each decomposition.

        """

        super().__init__(
            lo_bounds,
            up_bounds,
            father,
            level,
            id,
            children,
            score,
            seed=seed,
            n_seeds=n_seeds,
        )

        self.dim = len(self.up_bounds)
        self.n_dim = 2 * self.dim
        self.spokes = spokes

    def create_children(self):

        logger.info(f"Creating children of n°{self.id}")

        self.next_seeds = np.random.uniform(
            self.lo_bounds, self.up_bounds, (self.n_seeds, self.dim)
        )

        for i in self.next_seeds:
            child = BoxedVoronoi(
                self.lo_bounds,
                self.up_bounds,
                self,
                self.level + 1,
                len(self.children),
                seed=i,
                spokes=self.spokes,
                n_seeds=self.n_seeds,
            )
            self.children.append(child)

        for i in range(len(self.children) - 1):
            for j in range(i + 1, len(self.children)):
                try:
                    h = Hyperplane(self.children[i], self.children[j])
                    self.children[i].hyperplanes.append(h)
                    self.children[j].hyperplanes.append(h)
                except AssertionError as e:
                    logger.warning(f"Hyperplane building aborted: {e}")

        for i, c in enumerate(self.children):
            logger.info(f"Building children n°{i}/{len(self.children)}")
            if len(c.hyperplanes) > 0:
                c.update()
            else:
                self.children.pop(i)

    def update(self):

        sampled_bounds = []

        upma = self.up_bounds - self.seed
        loma = self.lo_bounds - self.seed

        cell_idx = [False] * len(self.hyperplanes)

        # Fixed points (2 for each dimension)
        for d in range(self.n_dim):

            l = self.dimSpoke(self.seed, d % self.dim)

            inter = np.empty((0, self.dim))

            for j, h in enumerate(self.hyperplanes):

                on, pfar = h.intersection(l)

                if on:
                    # Clip line to bounds
                    if np.any(pfar > self.up_bounds) or np.any(
                        pfar < self.lo_bounds
                    ):

                        pfar_clipped = self.clipBorder(l, upma, loma)
                        inter = np.append(inter, [pfar_clipped], axis=0)

                        cell_idx[j] = False

                    else:
                        inter = np.append(inter, [pfar], axis=0)

                        cell_idx[j] = True
                else:
                    cell_idx[j] = False

            if len(inter) == 0:
                pfar_clipped = self.clipBorder(l, upma, loma)
                sampled_bounds.append(np.copy(pfar_clipped))

            else:
                dist = np.linalg.norm(inter - l.A, axis=1)
                minidx = np.argmin(dist)
                a = np.copy(inter[minidx])
                sampled_bounds.append(a)

        # Random points
        for d in range(self.spokes):

            l = self.randomSpoke(self.seed)

            inter = np.empty((0, self.dim))

            for j, h in enumerate(self.hyperplanes):

                on, pfar = h.intersection(l)

                if on:
                    # Clip line to bounds
                    if np.any(pfar > self.up_bounds) or np.any(
                        pfar < self.lo_bounds
                    ):

                        pfar_clipped = self.clipBorder(l, upma, loma)
                        inter = np.append(inter, [pfar_clipped], axis=0)

                        cell_idx[j] = False

                    else:
                        inter = np.append(inter, [pfar], axis=0)

                        cell_idx[j] = True
                else:
                    cell_idx[j] = False

            if len(inter) == 0:
                pfar_clipped = self.clipBorder(l, upma, loma)
                sampled_bounds.append(np.copy(pfar_clipped))

            else:
                dist = np.linalg.norm(inter - l.A, axis=1)
                minidx = np.argmin(dist)
                a = np.copy(inter[minidx])
                sampled_bounds.append(a)

        self.lo_bounds = np.nanmin(sampled_bounds, axis=0)
        self.up_bounds = np.nanmax(sampled_bounds, axis=0)
        del self.hyperplanes



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

docs_extras = [
    "Sphinx >= 3.0.0",  # Force RTD to use >= 3.0.0
    "docutils",
    "pylons-sphinx-themes >= 1.0.8",  # Ethical Ads
    "pylons_sphinx_latesturl",
    "repoze.sphinx.autointerface",
    "sphinxcontrib-autoprogram",
    "sphinx-copybutton",
    "sphinx-tabs",
    "sphinx-panels",
    "sphinx-rtd-theme",
    "pillow>=6.2.0",
]

setuptools.setup(
    name="zellij",
    version="0.0.4",
    author="Thomas Firmin",
    author_email="thomas.firmin@univ-lille.fr",
    description="A software framework for HyperParameters Optimization",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: CeCILL-C Free Software License Agreement (CECILL-C)",
        "Operating System :: OS Independent",
    ],
    keywords=[
        "fractal",
        "continuous optimization",
        "global optimization",
        "black-box functions",
        "decision space partitioning",
        "exploration",
        "exploitation",
        "metaheuristics",
        "tree search",
    ],
    url="https://github.com/ThomasFirmin/zellij",
    project_urls={
        "Bug Tracker": "https://github.com/ThomasFirmin/zellij/issues",
    },
    package_dir={"": "lib"},
    packages=setuptools.find_packages("lib"),
    install_requires=[
        "numpy>=1.21.4",
        "DEAP>=1.3.1",
        "botorch>=0.6.3.1",
        "gpytorch>=1.6.0",
        "matplotlib>=3.5.0",
        "seaborn>=0.11.2",
        "pandas>=1.3.4",
    ],
    extras_require={"mpi": ["mpi4py>=3.1.2"], "docs": docs_extras},
    python_requires=">=3.6",
)



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:44:44+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


import numpy as np
from abc import abstractmethod
import matplotlib.pyplot as plt


class Cooling(object):
    """Cooling

    Cooling is a base object which define what a cooling Schedule is.

    Attributes
    ----------

    Tcurrent : float
        Current temperature

    cross : int
        Count the number of times Tend is crossed.

    T0 : float
        Initial temperature of the cooling schedule.\
         Higher temperature leads to higher acceptance of a worse solution. (more exploration)

    Tend : float
        Temperature threshold. When reached the temperature is violently increased proportionally to\
        <T0>. It allows to periodically easily escape from local optima.

    peaks : int, default=1
        Maximum number of crossed threshold according to `Tend`. The temperature will be increased\
        <peaks> times.

    Methods
    -------

    cool()
        Decrease temperature and return the current temperature.

    reset()
        Reset cooling schedule

    iterations()
        Get the theoretical number of iterations to end the schedule.

    """

    def __init__(self, T0, Tend, peaks=1):

        ##############
        # PARAMETERS #
        ##############

        self.T0 = T0
        self.Tend = Tend
        self.peaks = peaks

        #############
        # VARIABLES #
        #############
        self.Tcurrent = self.T0
        self.k = 0
        self.cross = 0

    @abstractmethod
    def cool(self):
        pass

    @abstractmethod
    def iterations(self):
        pass

    def reset(self):
        self.Tcurrent = self.T0
        self.k = 0
        self.cross = 0

    def show(self, filepath=""):

        pts = [self.cool() for i in range(self.iterations())]
        self.reset()

        fig, ax = plt.subplots(figsize=(19.2, 14.4))
        fig.suptitle(f"{self.__class__.__name__} cooling schedule")
        plt.xlabel("Iterations")
        plt.ylabel("Temperature")
        plt.plot(pts, ls="-", color="orange")

        if filepath:
            save_path = os.path.join(
                self.loss_func.plots_path, f"cooling_sa.png"
            )

            plt.savefig(save_path, bbox_inches="tight")
            plt.close()
        else:
            plt.show()
            plt.close()


class MulExponential(Cooling):
    """MulExponential

    Exponential multiplicative monotonic cooling.

    :math:`T_k = T_0.\\alpha^k`

    Attributes
    ----------
    alpha : float
        Decrease factor. 0.8<=`alpha`<=0.9
    T0 : float
        Initial temperature of the cooling schedule.\
         Higher temperature leads to higher acceptance of a worse solution. (more exploration)

    Tend : float
        Temperature threshold. When reached the temperature is violently increased proportionally to\
        `T0`. It allows to periodically easily escape from local optima.

    peaks : int, default=1
        Maximum number of crossed threshold according to `Tend`. The temperature will be increased\
        `peaks` times.

    Methods
    -------

    cool()
        Decrease temperature and return the current temperature.

    reset()
        Reset cooling schedule

    iterations()
        Get the theoretical number of iterations to end the schedule.

    """

    def __init__(self, alpha, T0, Tend, peaks=1):
        super().__init__(T0, Tend, peaks)
        self.alpha = alpha

    def cool(self):

        self.Tcurrent = self.T0 * self.alpha ** self.k

        if self.Tcurrent <= self.Tend:
            self.cross += 1
            self.k = 0
        else:
            self.k += 1

        return self.Tcurrent if self.cross < self.peaks else False

    def iterations(self):
        return (
            int(np.ceil(np.log(self.Tend / self.T0) / np.log(self.alpha)))
            * self.peaks
        )


class MulLogarithmic(Cooling):
    """MulLogarithmic

    Logarithmic multiplicative monotonic cooling.

    :math:`T_k = \\frac{T_0}{1+\\alpha.log(1+k)}`

    Parameters
    ----------
    alpha : float
        Decrease factor. `alpha`>1
    T0 : float
        Initial temperature of the cooling schedule.\
         Higher temperature leads to higher acceptance of a worse solution. (more exploration)

    Tend : float
        Temperature threshold. When reached the temperature is violently increased proportionally to\
        `T0`. It allows to periodically easily escape from local optima.

    peaks : int, default=1
        Maximum number of crossed threshold according to `Tend`. The temperature will be increased\
        `peaks` times.

    Methods
    -------

    cool()
        Decrease temperature and return the current temperature.

    reset()
        Reset cooling schedule

    iterations()
        Get the theoretical number of iterations to end the schedule.

    """

    def __init__(self, alpha, T0, Tend, peaks=1):
        super().__init__(T0, Tend, peaks)
        self.alpha = alpha

    def cool(self):

        self.Tcurrent = self.T0 / (1 + self.alpha * np.log(1 + self.k))

        if self.Tcurrent <= self.Tend:
            self.cross += 1
            self.k = 0
        else:
            self.k += 1

        return self.Tcurrent if self.cross < self.peaks else False

    def iterations(self):
        return (
            int(np.ceil(np.exp((self.T0 / self.Tend - 1 / self.alpha)) - 1))
            * self.peaks
        )


class MulLinear(Cooling):
    """MulLinear

    Linear multiplicative monotonic cooling.

    :math:`T_k = \\frac{T_0}{1+\\alpha.k}`

    Parameters
    ----------
    alpha : float
        Decrease factor. `alpha`>0
    T0 : float
        Initial temperature of the cooling schedule.\
         Higher temperature leads to higher acceptance of a worse solution. (more exploration)

    Tend : float
        Temperature threshold. When reached the temperature is violently increased proportionally to\
        `T0`. It allows to periodically easily escape from local optima.

    peaks : int, default=1
        Maximum number of crossed threshold according to `Tend`. The temperature will be increased\
        `peaks` times.

    Methods
    -------

    cool()
        Decrease temperature and return the current temperature.

    reset()
        Reset cooling schedule

    iterations()
        Get the theoretical number of iterations to end the schedule.

    """

    def __init__(self, alpha, T0, Tend, peaks=1):
        super().__init__(T0, Tend, peaks)
        self.alpha = alpha

    def cool(self):

        self.Tcurrent = self.T0 / (1 + self.alpha * self.k)

        if self.Tcurrent <= self.Tend:
            self.cross += 1
            self.k = 0
        else:
            self.k += 1

        return self.Tcurrent if self.cross < self.peaks else False

    def iterations(self):
        return int(np.ceil(self.T0 / (self.Tend * self.alpha))) * self.peaks


class MulQuadratic(Cooling):
    """MulQuadratic

    Quadratic multiplicative monotonic cooling.

    :math:`T_k = \\frac{T_0}{1+\\alpha.k^2}`

    Parameters
    ----------
    alpha : float
        Decrease factor. `alpha`>0
    T0 : float
        Initial temperature of the cooling schedule.\
         Higher temperature leads to higher acceptance of a worse solution. (more exploration)

    Tend : float
        Temperature threshold. When reached the temperature is violently increased proportionally to\
        `T0`. It allows to periodically easily escape from local optima.

    peaks : int, default=1
        Maximum number of crossed threshold according to `Tend`. The temperature will be increased\
        `peaks` times.

    Methods
    -------

    cool()
        Decrease temperature and return the current temperature.

    reset()
        Reset cooling schedule

    iterations()
        Get the theoretical number of iterations to end the schedule.

    """

    def __init__(self, alpha, T0, Tend, peaks=1):
        super().__init__(T0, Tend, peaks)
        self.alpha = alpha

    def cool(self):

        self.Tcurrent = self.T0 / (1 + self.alpha * self.k ** 2)

        if self.Tcurrent <= self.Tend:
            self.cross += 1
            self.k = 0
        else:
            self.k += 1

        return self.Tcurrent if self.cross < self.peaks else False

    def iterations(self):
        return (
            int(np.ceil(np.sqrt(self.T0 / (self.Tend * self.alpha))))
            * self.peaks
        )


class AddLinear(Cooling):
    """AddLinear

    Linear additive monotonic cooling.

    :math:`T_k = T_{end} + (T_0-T_{end})\\left(\\frac{cycles-k}{cycles}\\right)`

    Parameters
    ----------
    cycles : int
        Number of cooling cycles.

    T0 : float
        Initial temperature of the cooling schedule.\
         Higher temperature leads to higher acceptance of a worse solution. (more exploration)

    Tend : float
        Temperature threshold. When reached the temperature is violently increased proportionally to\
        `T0`. It allows to periodically easily escape from local optima.

    peaks : int, default=1
        Maximum number of crossed threshold according to `Tend`. The temperature will be increased\
        `peaks` times.

    Methods
    -------

    cool()
        Decrease temperature and return the current temperature.

    reset()
        Reset cooling schedule

    iterations()
        Get the theoretical number of iterations to end the schedule.

    """

    def __init__(self, cycles, T0, Tend, peaks=1):
        super().__init__(T0, Tend, peaks)
        self.cycles = cycles

    def cool(self):

        self.Tcurrent = self.Tend + (self.T0 - self.Tend) * (
            (self.cycles - self.k) / self.cycles
        )

        if self.k == self.cycles:
            self.cross += 1
            self.k = 0
        else:
            self.k += 1

        return self.Tcurrent if self.cross < self.peaks else False

    def iterations(self):
        return self.cycles * self.peaks


class AddQuadratic(Cooling):
    """AddQuadratic

    Quadratic additive monotonic cooling.

    :math:`T_k = T_{end} + (T_0-T_{end})\\left(\\frac{cycles-k}{cycles}\\right)^2`

    Attributes
    ----------
    cycles : int
        Number of cooling cycles.

    T0 : float
        Initial temperature of the cooling schedule.\
         Higher temperature leads to higher acceptance of a worse solution. (more exploration)

    Tend : float
        Temperature threshold. When reached the temperature is violently increased proportionally to\
        `T0`. It allows to periodically easily escape from local optima.

    peaks : int, default=1
        Maximum number of crossed threshold according to `Tend`. The temperature will be increased\
        `peaks` times.

    Methods
    -------

    cool()
        Decrease temperature and return the current temperature.

    reset()
        Reset cooling schedule

    iterations()
        Get the theoretical number of iterations to end the schedule.

    """

    def __init__(self, cycles, T0, Tend, peaks=1):
        super().__init__(T0, Tend, peaks)
        self.cycles = cycles

    def cool(self):

        self.Tcurrent = (
            self.Tend
            + (self.T0 - self.Tend)
            * ((self.cycles - self.k) / self.cycles) ** 2
        )

        if self.k == self.cycles:
            self.cross += 1
            self.k = 0
        else:
            self.k += 1

        return self.Tcurrent if self.cross < self.peaks else False

    def iterations(self):
        return self.cycles * self.peaks


class AddExponential(Cooling):
    """AddExponential

    Exponential additive monotonic cooling.

    :math:`T_k = T_{end} + \\frac{T_0-T_{end}}{1+e^{\\frac{2ln(T_0-T_{end})}{cycles}}(k-0,5cycles)}`

    Attributes
    ----------
    cycles : int
        Number of cooling cycles.

    T0 : float
        Initial temperature of the cooling schedule.\
         Higher temperature leads to higher acceptance of a worse solution. (more exploration)

    Tend : float
        Temperature threshold. When reached the temperature is violently increased proportionally to\
        `T0`. It allows to periodically easily escape from local optima.

    peaks : int, default=1
        Maximum number of crossed threshold according to `Tend`. The temperature will be increased\
        `peaks` times.

    Methods
    -------

    cool()
        Decrease temperature and return the current temperature.

    reset()
        Reset cooling schedule

    iterations()
        Get the theoretical number of iterations to end the schedule.

    """

    def __init__(self, cycles, T0, Tend, peaks=1):
        super().__init__(T0, Tend, peaks)
        self.cycles = cycles

    def cool(self):

        self.Tcurrent = self.Tend + (self.T0 - self.Tend) * (
            1
            / (
                1
                + np.exp(
                    2
                    * np.log(self.T0 - self.Tend)
                    / self.cycles
                    * (self.k - 0.5 * self.cycles)
                )
            )
        )

        if self.k == self.cycles:
            self.cross += 1
            self.k = 0
        else:
            self.k += 1

        return self.Tcurrent if self.cross < self.peaks else False

    def iterations(self):
        return self.cycles * self.peaks


class AddTrigonometric(Cooling):
    """AddTrigonometric

    Trigonometric additive monotonic cooling.

    :math:`T_k = T_{end} + 0,5(T_0-T_{end})(1+cos(\\frac{k.\\pi}{cycles}))`

    Attributes
    ----------
    cycles : int
        Number of cooling cycles.

    T0 : float
        Initial temperature of the cooling schedule.\
         Higher temperature leads to higher acceptance of a worse solution. (more exploration)

    Tend : float
        Temperature threshold. When reached the temperature is violently increased proportionally to\
        `T0`. It allows to periodically easily escape from local optima.

    peaks : int, default=1
        Maximum number of crossed threshold according to `Tend`. The temperature will be increased\
        `peaks` times.

    Methods
    -------

    cool()
        Decrease temperature and return the current temperature.

    reset()
        Reset cooling schedule

    iterations()
        Get the theoretical number of iterations to end the schedule.

    """

    def __init__(self, cycles, T0, Tend, peaks=1):
        super().__init__(T0, Tend, peaks)
        self.cycles = cycles

    def cool(self):

        self.Tcurrent = self.Tend + 0.5 * (self.T0 - self.Tend) * (
            1 + np.cos(self.k * np.pi / self.cycles)
        )

        if self.k == self.cycles:
            self.cross += 1
            self.k = 0
        else:
            self.k += 1

        return self.Tcurrent if self.cross < self.peaks else False

    def iterations(self):
        return self.cycles * self.peaks



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:45:08+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


from zellij.core.metaheuristic import Metaheuristic
from zellij.core.loss_func import FDA_loss_func

import numpy as np
import copy

import logging

logger = logging.getLogger("zellij.FDA")


class FDA(Metaheuristic):

    """FDA

    Fractal Decomposition Algorithm (FDA) is composed of 5 part:
        –  Fractal decomposition : FDA uses hyper-spheres or hyper-cubes to decompose the search-space into smaller sub-spaces in a fractal way.
        –  Tree search algorithm : Fractals form a tree, so FDA is also a tree search problem. It can use Best First Search, Beam Search or others algorithms from the A* family.
        –  Exploration : To explore a fractal, FDA requires an exploration algorithm, for example GA,or in our case CGS.
        –  Exploitation : At the final fractal level (e.g. a leaf of the rooted tree) FDA performs an exploitation.
        –  Scoring method: To score a fractal, FDA can use the best score found, the median, ... See heuristics.py.

    It a continuous optimization algorithm. SO the search space is converted to continuous.

    Attributes
    ----------

    heuristic : callable
        Determine using using current state of the algorithm, how to score the current fractal. Used informations are given to the function at the following order:
        - The current fractal
        - The best solution found so far (converted to continuous)
        - The best score found so far (computed with the loss function)

    exploration : Metaheuristic
        At each node of a fractal FDA applies an exploration algorithm to determine if this fractal is promising or not.

    exploitation : Metaheuristic
        At a leaf of the rooted fractal tree, FDA applies an exploitation algorithm, which ignores subspace bounds (not SearchSpace bounds),
        to refine the best solution found inside this fractal.

    level : int
        Depth of the fractal tree

    up_bounds : list
        List of float containing the upper bounds of the search space converted to continuous

    lo_bounds : list
        List of float containing the lower bounds of the search space converted to continuous

    fractal : Fractal
        Fractal object used to build the fractal tree

    explor_kwargs : list[dict]
        List of keyword arguments to pass to the exploration strategy at each level of the tree.
        If len(explor_kwargs) < level, then that last element of the list will be used for the next levels.

    explor_kwargs : dict
        Keyword arguments to pass to the exploitation strategy.

    start_H : Fractal
        Root of the fractal tree

    tree_search : Tree_search
        Tree_search object to use to explore and exploit the fractal tree.

    n_h : int
        Number of explored nodes of the tree

    total_h : int
        Theoretical number of nodes.


    Methods
    -------

    evaluate(hypervolumes)
        Evaluate a list of fractals using exploration and/or exploitation

    run(n_process=1)
        Runs FDA

    See Also
    --------
    Metaheuristic : Parent class defining what a Metaheuristic is
    LossFunc : Describes what a loss function is in Zellij
    Searchspace : Describes what a search space is in Zellij
    Tree_search : Tree search algorithm to explore and exploit the fractal tree.
    Fractal : Base class which defines what a fractal is.
    """

    def __init__(
        self,
        loss_func,
        search_space,
        f_calls,
        exploration,
        exploitation,
        fractal,
        tree_search,
        heuristic,
        level=5,
        verbose=True,
        **kwargs,
    ):

        """__init__(loss_func, search_space, f_calls, exploration, exploitation, fractal, tree_search, heuristic, level=5, verbose=True, **kwargs)

        Initialize FDA class

        Parameters
        ----------
        loss_func : Loss
            Loss function to optimize. must be of type f(x)=y

        search_space : Searchspace
            Search space object containing bounds of the search space.

        f_calls : int
            Maximum number of loss_func calls

        exploration : {Metaheuristic, list[Metaheuristic]}
            At each node of a fractal FDA applies an exploration algorithm to determine if this fractal is promising or not.
            If a list of metaheuristic is given, at each level FDA will use the metaheuristic at the index equel to the current level.
            If len(exploration) < level, the last metaheuristic will be used for following levels.

        exploitation : Metaheuristic
            At a leaf of the rooted fractal tree, FDA applies an exploitation algorithm, which ignores subspace bounds (not SearchSpace bounds),
            to refine the best solution found inside this fractal.

        level : int, default=5
            Fractal tree depth.

        tree_search : Tree_search
            BFS : Breadth first search
            DFS : Depth first search
            BS : Beam_search
            BestFS : Best first search
            CBFS : Cyclic best first search
            DBFS : Diverse best first search
            EGS : Epsilon greedy search

        fractal : Fractal
            Fractal used to build the fractal tree

        heuristic : callable
            Determine using using current state of the algorithm, how to score the current fractal. Used informations are given to the function at the following order:
            - The current fractal
            - The best solution found so far (converted to continuous)
            - The best score found so far (computed with the loss function)

        verbose : boolean, default=True
            Algorithm verbosity

        **kwargs : dict
            Keyword arguments for fractals and tree search algorithm.

        """

        ##############
        # PARAMETERS #
        ##############

        super().__init__(loss_func, search_space, f_calls, verbose)

        self.heuristic = heuristic
        self.level = level

        # Exploration and exploitation function
        if type(exploration) != list:
            self.exploration = [exploration]
        else:
            self.exploration = exploration

        if exploitation:
            self.exploitation = exploitation
            self.exploi_calls = self.exploitation.f_calls
        else:
            self.exploitation = False

        #############
        # VARIABLES #
        #############

        # Save f_calls from metaheuristic, to adapt them during FDA.
        self.explor_calls = [i.f_calls for i in self.exploration]

        # Working variables
        self.up_bounds = np.array([1.0 for _ in self.search_space.values])
        self.lo_bounds = np.array([0.0 for _ in self.search_space.values])

        self.fractal = fractal
        self.tree_search = tree_search

        self.volume_kwargs = {
            key: kwargs[key]
            for key in kwargs
            if key in self.fractal.__init__.__code__.co_varnames
        }
        self.ts_kwargs = {
            key: kwargs[key]
            for key in kwargs
            if key in self.tree_search.__init__.__code__.co_varnames
        }

        # Initialize first fractal
        self.root = self.fractal(
            self.lo_bounds, self.up_bounds, **self.volume_kwargs
        )

        # Initialize tree search
        self.tree_search = self.tree_search(
            [self.root], self.level, **self.ts_kwargs
        )

        # Number of explored hypersphere
        self.n_h = 0

        self.executed = False  # A voir

        # Best solution converted to continuous
        self.best_ind_c = []

        self.n_process = 1

    # Evaluate a list of hypervolumes
    def evaluate(self, hypervolumes):

        """evaluate(hypervolumes)

        Evaluate a list of fractals using exploration and/or exploitation.

        Parameters
        ----------
        hypervolumes : list[Fractal]
            list of hypervolumes to evaluate with exploration and/or exploitation

        """

        # While there are hypervolumes to evaluate do...
        i = 0
        while i < len(hypervolumes) and self.loss_func.calls < self.f_calls:

            # Select parent hypervolume
            H = hypervolumes[i]
            j = 0

            # While there are children do...
            while j < len(H.children) and self.loss_func.calls < self.f_calls:

                # Select children of parent H
                child = H.children[j]

                j += 1

                # Link the loss function to the actual hypervolume (children)
                modified_loss_func = FDA_loss_func(
                    self.loss_func, child, self.search_space
                )

                # Count the number of explored hypervolume
                self.n_h += 1

                # Exploration
                if child.level != self.level:

                    explor_idx = (
                        np.min([child.level, len(self.exploration)]) - 1
                    )
                    calls_left = np.min(
                        [
                            self.explor_calls[explor_idx],
                            self.f_calls - self.loss_func.calls,
                        ]
                    )

                    if calls_left > 0:

                        # Compute bounds of child hypervolume
                        lo = self.search_space.convert_to_continuous(
                            [child.lo_bounds], True, True
                        )[0]
                        up = self.search_space.convert_to_continuous(
                            [child.up_bounds], True, True
                        )[0]

                        # Create a search space for the metaheuristic
                        sp = self.search_space.subspace(lo, up)
                        self.exploration[explor_idx].search_space = sp
                        self.exploration[
                            explor_idx
                        ].loss_func = modified_loss_func

                        self.exploration[explor_idx].f_calls = (
                            calls_left + self.loss_func.calls
                        )

                        logger.info(
                            f"Exploration {self.fractal.__name__} n° {child.id} child of {child.father.id} at level {child.level}"
                        )
                        logger.info(
                            f"Explored {self.fractal.__name__}: {self.n_h}"
                        )

                        # Progress bar
                        prec_calls = self.loss_func.calls

                        # Run exploration, scores and evaluated solutions are saved using FDA_loss_func class
                        self.exploration[explor_idx].run(
                            H=child, n_process=self.n_process
                        )

                        # Save best found solution
                        if self.loss_func.new_best:
                            logger.info(
                                f"Best solution found :{self.loss_func.best_score}"
                            )
                            self.best_ind_c = (
                                self.search_space.convert_to_continuous(
                                    [self.loss_func.best_sol]
                                )[0]
                            )

                        child.score = self.heuristic(
                            child, self.best_ind_c, self.loss_func.best_score
                        )

                        logger.debug(
                            f"Child {child.father.id}.{child.id}.{child.level} score: {child.score}"
                        )

                        # Add child to tree search
                        self.tree_search.add(child)

                        # Progress bar
                        self.pending_pb(self.loss_func.calls - prec_calls)
                        self.update_main_pb(
                            self.loss_func.calls - prec_calls,
                            explor=True,
                            best=self.loss_func.new_best,
                        )
                        self.meta_pb.update(self.loss_func.calls - prec_calls)

                # Exploitation
                elif self.exploitation:

                    # Run exploitation, scores and evaluated solutions are saved using FDA_loss_func class
                    self.exploitation.loss_func = modified_loss_func
                    calls_left = np.min(
                        [self.exploi_calls, self.f_calls - self.loss_func.calls]
                    )

                    if calls_left > 0:

                        self.exploitation.f_calls = (
                            calls_left + self.loss_func.calls
                        )

                        logger.info(
                            f" -> Exploitation {self.fractal.__name__} n° {child.id} child of {child.father.id} at level {child.level}"
                        )
                        logger.info(
                            f"Explored {self.fractal.__name__}: {self.n_h}"
                        )

                        # Progress bar
                        prec_calls = self.loss_func.calls

                        # Run exploitation, scores and evaluated solutions are saved using FDA_loss_func class
                        self.exploitation.run(H=child, n_process=self.n_process)

                        if self.loss_func.new_best:
                            logger.info(
                                f"Best solution found :{self.loss_func.best_score}"
                            )
                            self.best_ind_c = (
                                self.search_space.convert_to_continuous(
                                    [self.loss_func.best_sol]
                                )[0]
                            )

                        logger.debug(
                            f"Child {child.father.id}.{child.id}.{child.level} score: EXPLOITaTION"
                        )

                        # Progress bar
                        self.pending_pb(self.loss_func.calls - prec_calls)
                        self.update_main_pb(
                            self.loss_func.calls - prec_calls,
                            explor=False,
                            best=self.loss_func.new_best,
                        )
                        self.meta_pb.update(self.loss_func.calls - prec_calls)

            i += 1

    def run(self, H=None, n_process=1):

        """run(H=None, n_process=1)

        Runs FDA.

        Parameters
        ----------
        H : Fractal, default=None
            When used by FDA, a fractal corresponding to the current subspace is given

        n_process : int, default=1
            Determine the number of best solution found to return.

        Returns
        -------
        best_sol : list[float]
            Returns a list of the <n_process> best found points to the continuous format

        best_scores : list[float]
            Returns a list of the <n_process> best found scores associated to best_sol

        """

        if H != None:
            logger.warning(f"Do not give any fractal to FDA.run")

        self.build_bar(self.f_calls)

        for exp in self.exploration:
            exp.manager = self.manager

        logger.info("Starting")

        if self.exploitation:
            self.exploitation.manager = self.manager

        self.n_process = n_process

        self.n_h = 0

        stop = True

        # Select initial hypervolume (root) from the search tree
        stop, hypervolumes = self.tree_search.get_next()

        while stop and self.loss_func.calls < self.f_calls:

            for H in hypervolumes:
                H.create_children()

            self.evaluate(hypervolumes)

            stop, hypervolumes = self.tree_search.get_next()

        self.executed = True

        logger.info(f"Loss function calls: {self.loss_func.calls}")
        logger.info(f"Explored {self.fractal.__name__}: {self.n_h}")
        logger.info(f"Best score: {self.loss_func.best_score}")
        logger.info(f"Best solution: {self.loss_func.best_sol}")

        best_idx = np.argpartition(self.loss_func.all_scores, n_process)
        best = [self.loss_func.all_solutions[i] for i in best_idx[:n_process]]
        min = [self.loss_func.all_scores[i] for i in best_idx[:n_process]]

        self.close_bar()

        logger.info("Ending")

        return best, min

    def show(self, filepath="", save=False):

        """show(filename=None)

        Plots solutions and scores evaluated during the optimization

        Parameters
        ----------
        filename : str, default=None
            If a filepath is given, the method will read the file and will try to plot contents.

        """

        super().show(filepath, save)
